"""Neighbor models for regression."""

from sklearn.neighbors import (
    KNeighborsRegressor as _KNeighborsRegressor,
    RadiusNeighborsRegressor as _RadiusNeighborsRegressor,
)

from .._neighbors_base import (
    KNeighborsMixin,
    NeighborsBase,
    NeighborsRegressorMixin,
    RadiusNeighborsMixin,
)


class KNeighborsRegressor(NeighborsBase, NeighborsRegressorMixin,
                          KNeighborsMixin):
    """Regression based on k-nearest neighbors.

    Regression with scalar, multivariate or functional response.

    The target is predicted by local interpolation of the targets associated of
    the nearest neighbors in the training set.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.
    weights : str or callable, optional (default = 'uniform')
        weight function used in prediction.  Possible values:

        - 'uniform' : uniform weights.  All points in each neighborhood
          are weighted equally.
        - 'distance' : weight points by the inverse of their distance.
          in this case, closer neighbors of a query point will have a
          greater influence than neighbors which are further away.
        - [callable] : a user-defined function which accepts an
          array of distances, and returns an array of the same shape
          containing the weights.

    regressor : callable, optional ((default =
        :func:`mean <skfda.exploratory.stats.mean>`))
        Function to perform the local regression in the functional response
        case. By default used the mean. Can the neighbors of a test sample,
        and if weights != 'uniform' an array of weights as second parameter.
    algorithm : {'auto', 'ball_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm based on
          the values passed to :meth:`fit` method.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.
    metric : string or callable, (default
        :func:`l2_distance <skfda.misc.metrics.l2_distance>`)
        the distance metric to use for the tree.  The default metric is
        the L2 distance. See the documentation of the metrics module
        for a list of available metrics.
    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.
    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        Doesn't affect :meth:`fit` method.
    multivariate_metric : boolean, optional (default = False)
        Indicates if the metric used is a sklearn distance between vectors (see
        :class:`sklearn.neighbors.DistanceMetric`) or a functional metric of
        the module :mod:`skfda.misc.metrics`.
    Examples
    --------
    Firstly, we will create a toy dataset with gaussian-like samples shifted.

    >>> from skfda.ml.regression import KNeighborsRegressor
    >>> from skfda.datasets import make_multimodal_samples
    >>> from skfda.datasets import make_multimodal_landmarks
    >>> y = make_multimodal_landmarks(n_samples=30, std=.5, random_state=0)
    >>> y_train = y.flatten()
    >>> X_train = make_multimodal_samples(n_samples=30, std=.5, random_state=0)
    >>> X_test = make_multimodal_samples(n_samples=5, std=.05, random_state=0)

    We will fit a K-Nearest Neighbors regressor to regress a scalar response.

    >>> neigh = KNeighborsRegressor()
    >>> neigh.fit(X_train, y_train)
    KNeighborsRegressor(...)

    We can predict the modes of new samples

    >>> neigh.predict(X_test).round(2) # Predict test data
    array([ 0.38, 0.14, 0.27, 0.52, 0.38])


    Now we will create a functional response to train the model

    >>> y_train = 5 * X_train + 1
    >>> y_train
    FDataGrid(...)

    We train the estimator with the functional response

    >>> neigh.fit(X_train, y_train)
    KNeighborsRegressor(...)

    And predict the responses as in the first case.

    >>> neigh.predict(X_test)
    FDataGrid(...)

    See also
    --------
    :class:`~skfda.ml.classification.KNeighborsClassifier`
    :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
    :class:`~skfda.ml.classification.NearestCentroids`
    :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
    :class:`~skfda.ml.clustering.NearestNeighbors`


    Notes
    -----
    See Nearest Neighbors in the sklearn online documentation for a discussion
    of the choice of ``algorithm`` and ``leaf_size``.

    This class wraps the sklearn regressor
    `sklearn.neighbors.KNeighborsRegressor`.

    .. warning::
       Regarding the Nearest Neighbors algorithms, if it is found that two
       neighbors, neighbor `k+1` and `k`, have identical distances
       but different labels, the results will depend on the ordering of the
       training data.

    https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    def __init__(self, n_neighbors=5, weights='uniform', regressor='mean',
                 algorithm='auto', leaf_size=30, metric='l2',
                 metric_params=None, n_jobs=1, multivariate_metric=False):
        """Initialize the regressor."""
        super().__init__(n_neighbors=n_neighbors,
                         weights=weights, algorithm=algorithm,
                         leaf_size=leaf_size, metric=metric,
                         metric_params=metric_params, n_jobs=n_jobs,
                         multivariate_metric=multivariate_metric)
        self.regressor = regressor

    def _init_multivariate_estimator(self, sklearn_metric):
        """Initialize the sklearn K neighbors estimator.

        Args:
            sklearn_metric: (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn K Neighbors estimator initialized.

        """
        return _KNeighborsRegressor(
            n_neighbors=self.n_neighbors, weights=self.weights,
            algorithm=self.algorithm, leaf_size=self.leaf_size,
            metric=sklearn_metric, metric_params=self.metric_params,
            n_jobs=self.n_jobs)

    def _query(self, X):
        """Return distances and neighbors of given sample."""
        return self.estimator_.kneighbors(X)


class RadiusNeighborsRegressor(NeighborsBase, NeighborsRegressorMixin,
                               RadiusNeighborsMixin):
    """Regression based on neighbors within a fixed radius.

    Regression with scalar, multivariate or functional response.

    The target is predicted by local interpolation of the targets associated of
    the nearest neighbors in the training set.

    Parameters
    ----------
    radius : float, optional (default = 1.0)
        Range of parameter space to use by default for :meth:`radius_neighbors`
        queries.
    weights : str or callable
        weight function used in prediction.  Possible values:

        - 'uniform' : uniform weights.  All points in each neighborhood
          are weighted equally.
        - 'distance' : weight points by the inverse of their distance.
          in this case, closer neighbors of a query point will have a
          greater influence than neighbors which are further away.
        - [callable] : a user-defined function which accepts an
          array of distances, and returns an array of the same shape
          containing the weights.

        Uniform weights are used by default.
    regressor : callable, optional ((default =
        :func:`mean <skfda.exploratory.stats.mean>`))
        Function to perform the local regression in the functional response
        case. By default used the mean. Can the neighbors of a test sample,
        and if weights != 'uniform' an array of weights as second parameter.
    algorithm : {'auto', 'ball_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree. This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree. The optimal value depends on the
        nature of the problem.
    metric : string or callable, (default
        :func:`l2_distance <skfda.metrics.l2_distance>`)
        the distance metric to use for the tree.  The default metric is
        the L2 distance. See the documentation of the metrics module
        for a list of available metrics.
    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.
    outlier_response : :class:`FData`, optional (default = None)
        Default response in the functional response case for test samples
        without neighbors.
    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
    multivariate_metric : boolean, optional (default = False)
        Indicates if the metric used is a sklearn distance between vectors (see
        :class:`sklearn.neighbors.DistanceMetric`) or a functional metric of
        the module :mod:`skfda.misc.metrics`.

    Examples
    --------
    Firstly, we will create a toy dataset with gaussian-like samples shifted.

    >>> from skfda.ml.regression import RadiusNeighborsRegressor
    >>> from skfda.datasets import make_multimodal_samples
    >>> from skfda.datasets import make_multimodal_landmarks
    >>> y = make_multimodal_landmarks(n_samples=30, std=.5, random_state=0)
    >>> y_train = y.flatten()
    >>> X_train = make_multimodal_samples(n_samples=30, std=.5, random_state=0)
    >>> X_test = make_multimodal_samples(n_samples=5, std=.05, random_state=0)

    We will fit a Radius-Nearest Neighbors regressor to regress a scalar
    response.

    >>> neigh = RadiusNeighborsRegressor(radius=0.2)
    >>> neigh.fit(X_train, y_train)
    RadiusNeighborsRegressor(...radius=0.2...)

    We can predict the modes of new samples

    >>> neigh.predict(X_test).round(2) # Predict test data
    array([ 0.39, 0.07, 0.26, 0.5 , 0.46])


    Now we will create a functional response to train the model

    >>> y_train = 5 * X_train + 1
    >>> y_train
    FDataGrid(...)

    We train the estimator with the functional response

    >>> neigh.fit(X_train, y_train)
    RadiusNeighborsRegressor(...radius=0.2...)

    And predict the responses as in the first case.

    >>> neigh.predict(X_test)
    FDataGrid(...)

    See also
    --------
    :class:`~skfda.ml.classification.KNeighborsClassifier`
    :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
    :class:`~skfda.ml.classification.NearestCentroids`
    :class:`~skfda.ml.regression.KNeighborsRegressor`
    :class:`~skfda.ml.clustering.NearestNeighbors`


    Notes
    -----
    See Nearest Neighbors in the sklearn online documentation for a discussion
    of the choice of ``algorithm`` and ``leaf_size``.

    This class wraps the sklearn classifier
    `sklearn.neighbors.RadiusNeighborsClassifier`.

    https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    def __init__(self, radius=1.0, weights='uniform', regressor='mean',
                 algorithm='auto', leaf_size=30, metric='l2',
                 metric_params=None, outlier_response=None, n_jobs=1,
                 multivariate_metric=False):
        """Initialize the classifier."""
        super().__init__(radius=radius, weights=weights, algorithm=algorithm,
                         leaf_size=leaf_size, metric=metric,
                         metric_params=metric_params, n_jobs=n_jobs,
                         multivariate_metric=multivariate_metric)
        self.regressor = regressor
        self.outlier_response = outlier_response

    def _init_multivariate_estimator(self, sklearn_metric):
        """Initialize the sklearn radius neighbors estimator.

        Args:
            sklearn_metric: (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn Radius Neighbors estimator initialized.

        """
        return _RadiusNeighborsRegressor(
            radius=self.radius, weights=self.weights,
            algorithm=self.algorithm, leaf_size=self.leaf_size,
            metric=sklearn_metric, metric_params=self.metric_params,
            n_jobs=self.n_jobs)

    def _query(self, X):
        """Return distances and neighbors of given sample.

        Args:
            X: the sample

        Returns:
            Distances and neighbors of a given sample

        """
        return self.estimator_.radius_neighbors(X)



"""Shift registration of functional data by least squares."""
from __future__ import annotations

import warnings
from typing import Callable, Optional, Tuple, TypeVar, Union

import numpy as np
from sklearn.utils.validation import check_is_fitted
from typing_extensions import Literal

from ... import FData, FDataGrid
from ..._utils import check_is_univariate
from ...misc._math import inner_product
from ...misc.metrics._lp_norms import l2_norm
from ...representation._typing import ArrayLike, GridPointsLike, NDArrayFloat
from ...representation.extrapolation import ExtrapolationLike
from .base import InductiveRegistrationTransformer

SelfType = TypeVar("SelfType", bound="LeastSquaresShiftRegistration[FData]")
T = TypeVar("T", bound=FData)
TemplateFunction = Callable[[FDataGrid], FDataGrid]


class LeastSquaresShiftRegistration(
    InductiveRegistrationTransformer[T, T],
):
    r"""Register data using shift alignment by least squares criterion.

    Realizes the registration of a set of curves using a shift aligment
    :footcite:`ramsay+silverman_2005_functional_shift`.
    Let :math:`\{x_i(t)\}_{i=1}^{N}` be a functional dataset, calculates
    :math:`\delta_{i}` for each sample such that :math:`x_i(t + \delta_{i})`
    minimizes the least squares criterion:

    .. math::
        \text{REGSSE} = \sum_{i=1}^{N} \int_{\mathcal{T}}
        [x_i(t + \delta_i) - \hat\mu(t)]^2 ds

    Estimates each shift parameter :math:`\delta_i` iteratively by
    using a modified Newton-Raphson algorithm, updating the template
    :math:`\mu` in each iteration as is described in detail in
    :footcite:`ramsay+silverman_2005_functional_newton-raphson`.

    Method only implemented for univariate functional data.

    Args:
        max_iter: Maximun number of iterations.
            Defaults sets to 5. Generally 2 or 3 iterations are sufficient to
            obtain a good alignment.
        tol: Tolerance allowable. The process will stop if
            :math:`\max_{i}|\delta_{i}^{(\nu)}-\delta_{i}^{(\nu-1)}|<tol`.
            Default sets to 1e-2.
        template: Template to use in the
            least squares criterion. If template="mean" it is use the
            functional mean as in the original paper. The template can be a
            callable that will receive an FDataGrid with the samples and will
            return another FDataGrid as a template, such as any of the means or
            medians of the module `skfda.explotatory.stats`.
            If the template is an FData is used directly as the final
            template to the registration, if it is a callable or "mean" the
            template is computed iteratively constructing a temporal template
            in each iteration.
            In :footcite:`ramsay+silverman_2005_functional_newton-raphson`
            is described in detail this procedure. Defaults to "mean".
        extrapolation: Controls the
            extrapolation mode for points outside the :term:`domain` range.
            By default uses the method defined in the data to be transformed.
            See the `extrapolation` documentation to obtain more information.
        step_size: Parameter to adjust the rate of
            convergence in the Newton-Raphson algorithm, see
            :footcite:`ramsay+silverman_2005_functional_newton-raphson`.
            Defaults to 1.
        restrict_domain: If True restricts the :term:`domain`
            to avoid the need of using extrapolation, in which
            case only the fit_transform method will be available, as training
            and transformation must be done together. Defaults to False.
        initial: Array with an initial estimation
            of shifts. Default uses a list of zeros for the initial shifts.
        grid_points: Set of points where the
            functions are evaluated to obtain the discrete
            representation of the object to integrate. If None is
            passed it calls numpy.linspace in FDataBasis and uses the
            `grid_points` in FDataGrids.

    Attributes:
        template\_: Template :math:`\mu` learned during the fitting
            used to the transformation.
        deltas\_: List of shifts :math:`\delta_i` applied
            during the last transformation.
        n_iter\_: Number of iterations performed during the last
            transformation.

    Note:
        Due to the use of derivatives for the estimation of the shifts, the
        samples to be registered may be smooth for the correct convergence of
        the method.

    Examples:
        >>> from skfda.preprocessing.registration import (
        ...     LeastSquaresShiftRegistration,
        ... )
        >>> from skfda.datasets import make_sinusoidal_process
        >>> from skfda.representation.basis import Fourier


        Registration and creation of dataset in discretized form:

        >>> fd = make_sinusoidal_process(n_samples=10, error_std=0,
        ...                              random_state=1)
        >>> reg = LeastSquaresShiftRegistration(extrapolation="periodic")
        >>> fd_registered = reg.fit_transform(fd)
        >>> fd_registered
        FDataGrid(...)

        Shifts applied during the transformation

        >>> reg.deltas_.round(3)
        array([-0.131,  0.188,  0.026,  0.033, -0.109,  0.115, ..., -0.062])


        Registration of a dataset in basis form using the
        transformation previosly fitted. The result is a dataset in
        discretized form, as it is not possible to express shifted functions
        exactly as a basis expansion:

        >>> fd = make_sinusoidal_process(n_samples=2, error_std=0,
        ...                              random_state=2)
        >>> fd_basis = fd.to_basis(Fourier())
        >>> reg.transform(fd_basis)
        FDataGrid(...)


    References:
        .. footbibliography::

    """

    def __init__(
        self,
        max_iter: int = 5,
        tol: float = 1e-2,
        template: Union[Literal["mean"], FData, TemplateFunction] = "mean",
        extrapolation: Optional[ExtrapolationLike] = None,
        step_size: float = 1,
        restrict_domain: bool = False,
        initial: Union[Literal["zeros"], ArrayLike] = "zeros",
        grid_points: Optional[GridPointsLike] = None,
    ) -> None:
        self.max_iter = max_iter
        self.tol = tol
        self.template = template
        self.restrict_domain = restrict_domain
        self.extrapolation = extrapolation
        self.step_size = step_size
        self.initial = initial
        self.grid_points = grid_points

    def _compute_deltas(
        self,
        fd: FData,
        template: Union[Literal["mean"], FData, TemplateFunction],
    ) -> Tuple[NDArrayFloat, FDataGrid]:
        """Compute the shifts to perform the registration.

        Args:
            fd: Functional object to be registered.
            template: Template to align the
                the samples. "mean" to compute the mean iteratively as in
                the original paper, an FData with the templated calculated or
                a callable wich constructs the template.

        Returns:
            A tuple with an array of deltas and an FDataGrid with the template.

        """
        check_is_univariate(fd)

        domain_range = fd.domain_range[0]

        # Initial estimation of the shifts
        if self.initial == "zeros":
            delta = np.zeros(fd.n_samples)
        else:
            delta = np.asarray(self.initial)

            if len(delta) != fd.n_samples:
                raise ValueError(
                    f"The length of the initial shift ({len(delta)}) must "
                    f"be the same than the number of samples ({fd.n_samples})",
                )

        # Auxiliar array to avoid multiple memory allocations
        delta_aux = np.empty(fd.n_samples)

        # Computes the derivate of originals curves in the mesh points
        fd_deriv = fd.derivative()

        # Second term of the second derivate estimation of REGSSE. The
        # first term has been dropped to improve convergence (see references)
        d2_regsse = l2_norm(fd_deriv)**2

        # We need the discretized derivative to compute the inner product later
        fd_deriv = fd_deriv.to_grid(grid_points=self.grid_points)

        max_diff = self.tol + 1
        self.n_iter_ = 0

        # Newton-Rhapson iteration
        while max_diff > self.tol and self.n_iter_ < self.max_iter:

            # Computes the new values shifted
            x = fd.shift(delta, grid_points=self.grid_points)

            if isinstance(template, str):
                assert template == "mean"
                template_iter = x.mean()
            elif isinstance(template, FData):
                template_iter = template.to_grid(grid_points=x.grid_points)
            else:  # Callable
                template_iter = template(x)

            # Updates the limits for non periodic functions ignoring the ends
            if self.restrict_domain:
                # Calculates the new limits
                a = domain_range[0] - min(np.min(delta), 0)
                b = domain_range[1] - max(np.max(delta), 0)

                restricted_domain = (
                    max(a, template_iter.domain_range[0][0]),
                    min(b, template_iter.domain_range[0][1]),
                )

                template_iter = template_iter.restrict(restricted_domain)

                x = x.restrict(restricted_domain)
                fd_deriv = fd_deriv.restrict(restricted_domain)
                d2_regsse = l2_norm(fd_deriv)**2

            # Calculates x - mean
            x -= template_iter

            d1_regsse = inner_product(x, fd_deriv)

            # Updates the shifts by the Newton-Rhapson iteration
            # Same as delta = delta - step_size * d1_regsse / d2_regsse
            delta_aux[:] = d1_regsse
            delta_aux[:] /= d2_regsse
            delta_aux[:] *= self.step_size
            delta[:] -= delta_aux

            # Updates convergence criterions
            max_diff = np.abs(delta_aux, out=delta_aux).max()
            self.n_iter_ += 1

        return delta, template_iter

    def fit_transform(self, X: T, y: None = None) -> T:

        deltas, template = self._compute_deltas(X, self.template)

        self.deltas_ = deltas
        self.template_ = template

        shifted = X.shift(
            self.deltas_,
            restrict_domain=self.restrict_domain,
            extrapolation=self.extrapolation,
            grid_points=self.grid_points,
        )
        shifted.argument_names = None
        return shifted

    def fit(
        self: SelfType,
        X: FData,
        y: None = None,
    ) -> SelfType:

        # If the template is an FData, fit doesnt learn anything
        if isinstance(self.template, FData):
            self.template_ = self.template

        else:
            _, template = self._compute_deltas(X, self.template)

            self.template_ = template

        return self

    def transform(self, X: FData, y: None = None) -> FDataGrid:

        if self.restrict_domain:
            raise AttributeError(
                "transform is not available when "
                "restrict_domain=True, fitting and "
                "transformation should be done together. Use "
                "an extrapolation method with "
                "restrict_domain=False or fit_predict",
            )

        # Check is fitted
        check_is_fitted(self)

        deltas, _ = self._compute_deltas(X, self.template_)
        self.deltas_ = deltas

        shifted = X.shift(
            deltas,
            restrict_domain=self.restrict_domain,
            extrapolation=self.extrapolation,
            grid_points=self.grid_points,
        )
        shifted.argument_names = None
        return shifted

    def inverse_transform(self, X: FData, y: None = None) -> FDataGrid:
        """
        Apply the inverse transformation.

        Applies the opossite shift used in the last call to `transform`.

        Args:
            X: Functional dataset to be transformed.
            y: not used, present for API consistency by convention.

        Returns:
            Functional data registered.

        Examples:
            Creates a synthetic functional dataset.

            >>> from skfda.preprocessing.registration import (
            ...     LeastSquaresShiftRegistration,
            ... )
            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd = make_sinusoidal_process(error_std=0, random_state=1)
            >>> fd.extrapolation = 'periodic'

            Dataset registration and centering.

            >>> reg = LeastSquaresShiftRegistration()
            >>> fd_registered = reg.fit_transform(fd)
            >>> fd_centered = fd_registered - fd_registered.mean()

            Reverse the translation applied during the registration.

            >>> reg.inverse_transform(fd_centered)
            FDataGrid(...)

        """
        deltas = getattr(self, "deltas_", None)

        if deltas is None:
            raise AttributeError(
                "Data must be previously transformed to learn"
                " the inverse transformation",
            )
        elif len(X) != len(deltas):
            raise ValueError(
                "Data must contain the same number of samples "
                "than the dataset previously transformed",
            )

        return X.shift(
            -deltas,
            restrict_domain=self.restrict_domain,
            extrapolation=self.extrapolation,
            grid_points=self.grid_points,
        )


class ShiftRegistration(LeastSquaresShiftRegistration[T]):
    """Deprecated name for LeastSquaresShiftRegistration."""

    def __init__(
        self,
        max_iter: int = 5,
        tol: float = 1e-2,
        template: Union[Literal["mean"], FData, TemplateFunction] = "mean",
        extrapolation: Optional[ExtrapolationLike] = None,
        step_size: float = 1,
        restrict_domain: bool = False,
        initial: Union[Literal["zeros"], ArrayLike] = "zeros",
        grid_points: Optional[GridPointsLike] = None,
    ) -> None:
        warnings.warn(
            "ShiftRegistration has been renamed. "
            "Use LeastSquaresShiftRegistration instead.",
            DeprecationWarning,
        )
        super().__init__(
            max_iter=max_iter,
            tol=tol,
            template=template,
            extrapolation=extrapolation,
            step_size=step_size,
            restrict_domain=restrict_domain,
            initial=initial,
            grid_points=grid_points,
        )



from . import maxima_hunting, mrmr, recursive_maxima_hunting
from ._rkvs import RKHSVariableSelection
from .maxima_hunting import MaximaHunting
from .mrmr import MinimumRedundancyMaximumRelevance
from .recursive_maxima_hunting import RecursiveMaximaHunting



"""
This module contains the structure of the evaluator.

The evaluator is the core of the FData object for extrapolation and
evaluation of FDataGrids.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Iterable, Union, overload

import numpy as np
from typing_extensions import Literal, Protocol

from ._typing import ArrayLike

if TYPE_CHECKING:
    from . import FData


class Evaluator(ABC):
    """
    Structure of an evaluator.

    An evaluator defines how to evaluate points of a functional object, it
    can be used as extrapolator to evaluate points outside the :term:`domain`
    range or as interpolation in a :class:`FDataGrid`. The corresponding
    examples of Interpolation and Extrapolation shows the basic usage of
    this class.

    The evaluator is called internally by :func:`evaluate`.

    """

    @abstractmethod
    def _evaluate(
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:
        """
        Evaluate the samples at evaluation points.

        Must be overriden in subclasses.

        """
        pass

    @overload
    def __call__(
        self,
        fdata: FData,
        eval_points: ArrayLike,
        *,
        aligned: Literal[True] = True,
    ) -> np.ndarray:
        pass

    @overload
    def __call__(
        self,
        fdata: FData,
        eval_points: Iterable[ArrayLike],
        *,
        aligned: Literal[False],
    ) -> np.ndarray:
        pass

    def __call__(
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:
        """
        Evaluate the samples at evaluation points.

        The evaluation call will receive a 2-d array with the
        evaluation points, or a 3-d array with the evaluation points per
        sample if ``aligned`` is ``False``.

        Args:
            fdata: Object to evaluate.
            eval_points: Numpy array with shape
                ``(number_eval_points, dim_domain)`` with the
                evaluation points.
            aligned: Whether the input points are
                the same for each sample, or an array of points per sample is
                passed.

        Returns:
            (numpy.darray): Numpy 3d array with shape
                ``(n_samples, number_eval_points, dim_codomain)`` with the
                result of the evaluation. The entry ``(i,j,k)`` will contain
                the value k-th image dimension of the i-th sample, at the
                j-th evaluation point.

        """
        return self._evaluate(
            fdata=fdata,
            eval_points=eval_points,
            aligned=aligned,
        )

    def __repr__(self) -> str:
        return f"{type(self)}()"

    def __eq__(self, other: Any) -> bool:
        """Equality operator between evaluators."""
        return isinstance(other, type(self))


class EvaluateFunction(Protocol):
    """Callback of an evaluation function."""

    def __call__(
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:
        """
        Evaluate the samples at evaluation points.

        The evaluation call will receive a 2-d array with the
        evaluation points, or a 3-d array with the evaluation points per
        sample if ``aligned`` is ``False``.

        Args:
            fdata: Object to evaluate.
            eval_points: Numpy array with shape
                ``(number_eval_points, dim_domain)`` with the
                evaluation points.
            aligned: Whether the input points are
                the same for each sample, or an array of points per sample is
                passed.

        Returns:
            Numpy 3d array with shape
            ``(n_samples, number_eval_points, dim_codomain)`` with the
            result of the evaluation. The entry ``(i,j,k)`` will contain
            the value k-th image dimension of the i-th sample, at the
            j-th evaluation point.

        """
        pass


class GenericEvaluator(Evaluator):
    """Generic Evaluator.

    Generic evaluator that recibes a functions to construct the evaluator.
    The function will recieve an :class:`FData` as first argument, a numpy
    array with the eval_points and the ``aligned`` parameter.

    """

    def __init__(self, evaluate_function: EvaluateFunction) -> None:
        self.evaluate_function = evaluate_function

    def _evaluate(  # noqa: D102
        self,
        fdata: FData,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        return self.evaluate_function(fdata, eval_points, aligned=aligned)



"""Base classes for the neighbor estimators"""

from abc import ABC

import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from .. import FData, FDataGrid
from ..misc.metrics import l2_distance


def _to_multivariate(fdatagrid):
    r"""Returns the data matrix of a fdatagrid in flatten form compatible with
    sklearn.

    Args:
        fdatagrid (:class:`FDataGrid`): Grid to be converted to matrix

    Returns:
        (np.array): Numpy array with size (n_samples, points), where
            points = prod([len(d) for d in fdatagrid.grid_points]

    """
    return fdatagrid.data_matrix.reshape(fdatagrid.n_samples, -1)


def _from_multivariate(data_matrix, grid_points, shape, **kwargs):
    r"""Constructs a FDatagrid from the data matrix flattened.

    Args:
        data_matrix (np.array): Data Matrix flattened as multivariate vector
            compatible with sklearn.
        grid_points (array_like): List with sample points for each dimension.
        shape (tuple): Shape of the data_matrix.
        **kwargs: Named params to be passed to the FDataGrid constructor.

    Returns:
        (:class:`FDataGrid`): FDatagrid with the data.

    """
    return FDataGrid(data_matrix.reshape(shape), grid_points, **kwargs)


def _to_multivariate_metric(metric, grid_points):
    r"""Transform a metric between FDatagrid in a sklearn compatible one.

    Given a metric between FDatagrids returns a compatible metric used to
    wrap the sklearn routines.

    Args:
        metric (pyfunc): Metric of the module `mics.metrics`. Must accept
            two FDataGrids and return a float representing the distance.
        grid_points (array_like): Array of arrays with the sample points of
            the FDataGrids.

    Returns:
        (pyfunc): sklearn vector metric.

    Examples:

        >>> import numpy as np
        >>> from skfda import FDataGrid
        >>> from skfda.misc.metrics import l2_distance
        >>> from skfda.ml._neighbors_base import _to_multivariate_metric

        Calculate the Lp distance between fd and fd2.

        >>> x = np.linspace(0, 1, 101)
        >>> fd = FDataGrid([np.ones(len(x))], x)
        >>> fd2 =  FDataGrid([np.zeros(len(x))], x)
        >>> l2_distance(fd, fd2).round(2)
        array([ 1.])

        Creation of the sklearn-style metric.

        >>> sklearn_l2_distance = _to_multivariate_metric(l2_distance, [x])
        >>> sklearn_l2_distance(np.ones(len(x)), np.zeros(len(x))).round(2)
        array([ 1.])

    """
    # Shape -> (n_samples = 1, domain_dims...., image_dimension (-1))
    shape = [1] + [len(axis) for axis in grid_points] + [-1]

    def multivariate_metric(x, y, **kwargs):

        return metric(_from_multivariate(x, grid_points, shape),
                      _from_multivariate(y, grid_points, shape),
                      **kwargs)

    return multivariate_metric


class NeighborsBase(ABC, BaseEstimator):
    """Base class for nearest neighbors estimators."""

    def __init__(self, n_neighbors=None, radius=None,
                 weights='uniform', algorithm='auto',
                 leaf_size=30, metric='l2', metric_params=None,
                 n_jobs=None, multivariate_metric=False):
        """Initializes the nearest neighbors estimator"""

        self.n_neighbors = n_neighbors
        self.radius = radius
        self.weights = weights
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.metric = metric
        self.metric_params = metric_params
        self.n_jobs = n_jobs
        self.multivariate_metric = multivariate_metric

    def _check_is_fitted(self):
        """Check if the estimator is fitted.

        Raises:
            NotFittedError: If the estimator is not fitted.

        """
        sklearn_check_is_fitted(self, ['estimator_'])

    def _transform_to_multivariate(self, X):
        """Transform the input data to array form. If the metric is
        precomputed it is not transformed.

        """
        if X is not None and self.metric != 'precomputed':
            X = _to_multivariate(X)

        return X


class NeighborsMixin:
    """Mixin class to train the neighbors models"""

    def fit(self, X, y=None):
        """Fit the model using X as training data and y as target values.

        Args:
            X (:class:`FDataGrid`, array_matrix): Training data. FDataGrid
                with the training data or array matrix with shape
                [n_samples, n_samples] if metric='precomputed'.
            y (array-like or sparse matrix): Target values of
                shape = [n_samples] or [n_samples, n_outputs].
                In the case of unsupervised search, this parameter is ignored.

        Note:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        # If metric is precomputed no diferences with the Sklearn stimator
        if self.metric == 'precomputed':
            self.estimator_ = self._init_estimator(self.metric)
            self.estimator_.fit(X, y)
        else:
            self._grid_points = X.grid_points
            self._shape = X.data_matrix.shape[1:]

            if not self.multivariate_metric:
                # Constructs sklearn metric to manage vector
                if self.metric == 'l2':
                    metric = l2_distance
                else:
                    metric = self.metric

                sklearn_metric = _to_multivariate_metric(metric,
                                                         self._grid_points)
            else:
                sklearn_metric = self.metric

            self.estimator_ = self._init_estimator(sklearn_metric)
            self.estimator_.fit(self._transform_to_multivariate(X), y)

        return self


class KNeighborsMixin:
    """Mixin class for K-Neighbors"""

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        """Finds the K-neighbors of a point.
        Returns indices of and distances to the neighbors of each point.

        Args:
            X (:class:`FDataGrid` or matrix): FDatagrid with the query
                functions or  matrix (n_query, n_indexed) if
                metric == 'precomputed'. If not provided, neighbors of each
                indexed point are returned. In this case, the query point is
                not considered its own neighbor.
            n_neighbors (int): Number of neighbors to get (default is the value
                passed to the constructor).
        return_distance (boolean, optional): Defaults to True. If False,
            distances will not be returned.

        Returns:
            dist : array
                Array representing the lengths to points, only present if
                return_distance=True
            ind : array
                Indices of the nearest points in the population matrix.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors()
            >>> neigh.fit(fd)
            NearestNeighbors(...)

            Now we can query the k-nearest neighbors.

            >>> distances, index = neigh.kneighbors(fd[:2])
            >>> index # Index of k-neighbors of samples 0 and 1
            array([[ 0,  7,  6, 11,  2],...)

            >>> distances.round(2) # Distances to k-neighbors
            array([[ 0.  ,  0.28,  0.29,  0.29,  0.3 ],
                   [ 0.  ,  0.27,  0.28,  0.29,  0.3 ]])

        Notes:
            This method wraps the corresponding sklearn routine in the
            module ``sklearn.neighbors``.

        """
        self._check_is_fitted()
        X = self._transform_to_multivariate(X)

        return self.estimator_.kneighbors(X, n_neighbors, return_distance)

    def kneighbors_graph(self, X=None, n_neighbors=None, mode='connectivity'):
        """Computes the (weighted) graph of k-Neighbors for points in X

        Args:
            X (:class:`FDataGrid` or matrix): FDatagrid with the query
                functions or  matrix (n_query, n_indexed) if
                metric == 'precomputed'. If not provided, neighbors of each
                indexed point are returned. In this case, the query point is
                not considered its own neighbor.
            n_neighbors (int): Number of neighbors to get (default is the value
                passed to the constructor).
            mode ('connectivity' or 'distance', optional): Type of returned
                matrix: 'connectivity' will return the connectivity matrix with
                ones and zeros, in 'distance' the edges are distance between
                points.

        Returns:
            Sparse matrix in CSR format, shape = [n_samples, n_samples_fit]
            n_samples_fit is the number of samples in the fitted data
            A[i, j] is assigned the weight of edge that connects i to j.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator.

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors()
            >>> neigh.fit(fd)
            NearestNeighbors(...)

            Now we can obtain the graph of k-neighbors of a sample.

            >>> graph = neigh.kneighbors_graph(fd[0])
            >>> print(graph)
              (0, 0)	1.0
              (0, 7)	1.0
              (0, 6)	1.0
              (0, 11)	1.0
              (0, 2)	1.0

        Notes:
            This method wraps the corresponding sklearn routine in the
            module ``sklearn.neighbors``.

        """
        self._check_is_fitted()

        X = self._transform_to_multivariate(X)

        return self.estimator_.kneighbors_graph(X, n_neighbors, mode)


class RadiusNeighborsMixin:
    """Mixin Class for Raius Neighbors"""

    def radius_neighbors(self, X=None, radius=None, return_distance=True):
        """Finds the neighbors within a given radius of a fdatagrid or
        fdatagrids.
        Return the indices and distances of each point from the dataset
        lying in a ball with size ``radius`` around the points of the query
        array. Points lying on the boundary are included in the results.
        The result points are *not* necessarily sorted by distance to their
        query point.

        Args:
            X (:class:`FDataGrid`, optional): fdatagrid with the sample or
                samples whose neighbors will be returned. If not provided,
                neighbors of each indexed point are returned. In this case, the
                query point is not considered its own neighbor.
            radius (float, optional): Limiting distance of neighbors to return.
                (default is the value passed to the constructor).
            return_distance (boolean, optional). Defaults to True. If False,
                distances will not be returned

        Returns
            (array, shape (n_samples): dist : array of arrays representing the
                distances to each point, only present if return_distance=True.
                The distance values are computed according to the ``metric``
                constructor parameter.
            (array, shape (n_samples,): An array of arrays of indices of the
                approximate nearest points from the population matrix that lie
                within a ball of size ``radius`` around the query points.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator.

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors(radius=.3)
            >>> neigh.fit(fd)
            NearestNeighbors(...radius=0.3...)

            Now we can query the neighbors in the radius.

            >>> distances, index = neigh.radius_neighbors(fd[:2])
            >>> index[0] # Neighbors of sample 0
            array([ 0,  2,  6,  7, 11]...)

            >>> distances[0].round(2) # Distances to neighbors of the sample 0
            array([ 0.  ,  0.3 ,  0.29,  0.28,  0.29])


        See also:
            kneighbors

        Notes:

            Because the number of neighbors of each point is not necessarily
            equal, the results for multiple query points cannot be fit in a
            standard data array.
            For efficiency, `radius_neighbors` returns arrays of objects, where
            each object is a 1D array of indices or distances.

            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        self._check_is_fitted()

        X = self._transform_to_multivariate(X)

        return self.estimator_.radius_neighbors(
            X=X, radius=radius, return_distance=return_distance)

    def radius_neighbors_graph(self, X=None, radius=None, mode='connectivity'):
        """Computes the (weighted) graph of Neighbors for points in X
        Neighborhoods are restricted the points at a distance lower than
        radius.

        Args:
            X  (:class:`FDataGrid`):  The query sample or samples. If not
                provided, neighbors of each indexed point are returned. In this
                case, the query point is not considered its own neighbor.
            radius (float): Radius of neighborhoods. (default is the value
                passed to the constructor).
            mode ('connectivity' or 'distance', optional): Type of returned
                matrix: 'connectivity' will return the connectivity matrix with
                ones and zeros, in 'distance' the edges are distance between
                points.

        Returns:
            sparse matrix in CSR format, shape = [n_samples, n_samples]
            A[i, j] is assigned the weight of edge that connects i to j.

        Notes:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.
        """
        self._check_is_fitted()

        X = self._transform_to_multivariate(X)

        return self.estimator_.radius_neighbors_graph(X=X, radius=radius,
                                                      mode=mode)


class NeighborsClassifierMixin:
    """Mixin class for classifiers based in nearest neighbors"""

    def predict(self, X):
        """Predict the class labels for the provided data.

        Args:
            X (:class:`FDataGrid` or array-like): FDataGrid with the test
                samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns:

            (np.array): y : array of shape [n_samples] or
            [n_samples, n_outputs] with class labels for each data sample.

        Notes:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        self._check_is_fitted()

        X = self._transform_to_multivariate(X)

        return self.estimator_.predict(X)


class NeighborsRegressorMixin(NeighborsMixin, RegressorMixin):
    """Mixin class for the regressors based on neighbors"""

    def _mean_regressor(self, X, weights=None):
        """
        Default regressor using weighted average.

        """

        if weights is None:
            return X.mean()
        else:
            weights /= np.sum(weights)

            return (X * weights).sum()

    def fit(self, X, y):
        """Fit the model using X as training data and y as responses.

        Args:
            X (:class:`FDataGrid`, array_matrix): Training data. FDataGrid
                with the training data or array matrix with shape
                [n_samples, n_samples] if metric='precomputed'.
            Y (:class:`FData` or array_like): Training data. FData
                with the training respones (functional response case)
                or array matrix with length `n_samples` in the multivariate
                response case.
        Returns:
            Estimator: self.

        """
        self._functional = isinstance(y, FData)

        if self._functional:
            return self._functional_fit(X, y)
        else:
            return super().fit(X, y)

    def _functional_fit(self, X, y):
        """Fit the model using X as training data.

        Args:
            X (:class:`FDataGrid`, array_matrix): Training data. FDataGrid
                with the training data or array matrix with shape
                [n_samples, n_samples] if metric='precomputed'.


        """
        if len(X) != y.n_samples:
            raise ValueError("The response and dependent variable must "
                             "contain the same number of samples,")

        # If metric is precomputed no different with the Sklearn stimator
        if self.metric == 'precomputed':
            self.estimator_ = self._init_estimator(self.metric)
            self.estimator_.fit(X)
        else:
            self._grid_points = X.grid_points
            self._shape = X.data_matrix.shape[1:]

            if not self.multivariate_metric:

                if self.metric == 'l2':
                    metric = l2_distance
                else:
                    metric = self.metric

                sklearn_metric = _to_multivariate_metric(metric,
                                                         self._grid_points)
            else:
                sklearn_metric = self.metric

            self.estimator_ = self._init_estimator(sklearn_metric)
            self.estimator_.fit(self._transform_to_multivariate(X))

        if self.regressor == 'mean':
            self._regressor = self._mean_regressor
        else:
            self._regressor = self.regressor

        # Choose proper local regressor
        if self.weights == 'uniform':
            self._local_regressor = self._uniform_local_regression
        elif self.weights == 'distance':
            self._local_regressor = self._distance_local_regression
        else:
            self._local_regressor = self._weighted_local_regression

        # Store the responses
        self._y = y

        return self

    def _uniform_local_regression(self, neighbors, distance=None):
        """Perform local regression with uniform weights"""
        return self._regressor(neighbors)

    def _distance_local_regression(self, neighbors, distance):
        """Perform local regression using distances as weights"""
        idx = distance == 0.
        if np.any(idx):
            weights = distance
            weights[idx] = 1. / np.sum(idx)
            weights[~idx] = 0.
        else:
            weights = 1. / distance
            weights /= np.sum(weights)

        return self._regressor(neighbors, weights)

    def _weighted_local_regression(self, neighbors, distance):
        """Perform local regression using custom weights"""

        weights = self.weights(distance)

        return self._regressor(neighbors, weights)

    def predict(self, X):
        """Predict the target for the provided data

        Args:
            X (:class:`FDataGrid` or array-like): FDataGrid with the test
                samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns:
            y : array of shape = [n_samples] or [n_samples, n_outputs]
                or :class:`FData` containing as many samples as X.

        """
        self._check_is_fitted()

        # Choose type of prediction
        if self._functional:
            return self._functional_predict(X)
        else:
            return self._multivariate_predict(X)

    def _multivariate_predict(self, X):
        """Predict the target for the provided data
        Parameters
        ----------
        X (:class:`FDataGrid` or array-like): FDataGrid with the test
            samples or array (n_query, n_indexed) if metric ==
            'precomputed'.
        Returns
        -------
        y : array of int, shape = [n_samples] or [n_samples, n_outputs]
            Target values
        Notes
        -----
        This method wraps the corresponding sklearn routine in the module
        ``sklearn.neighbors``.

        """
        X = self._transform_to_multivariate(X)

        return self.estimator_.predict(X)

    def _init_estimator(self, sklearn_metric):
        """Initialize the sklearn nearest neighbors estimator.

        Args:
            sklearn_metric: (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn K Neighbors estimator initialized.

        """
        if self._functional:
            from sklearn.neighbors import NearestNeighbors as _NearestNeighbors

            return _NearestNeighbors(
                n_neighbors=self.n_neighbors, radius=self.radius,
                algorithm=self.algorithm, leaf_size=self.leaf_size,
                metric=sklearn_metric, metric_params=self.metric_params,
                n_jobs=self.n_jobs)
        else:
            return self._init_multivariate_estimator(sklearn_metric)

    def _functional_predict(self, X):
        """Predict functional responses.

        Args:
            X (:class:`FDataGrid` or array-like): FDataGrid with the test
                samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns

            y : :class:`FDataGrid` containing as many samples as X.

        """

        X = self._transform_to_multivariate(X)

        distances, neighbors = self._query(X)

        if len(neighbors[0]) == 0:
            pred = self._outlier_response(neighbors)
        else:
            pred = self._local_regressor(self._y[neighbors[0]], distances[0])

        for i, idx in enumerate(neighbors[1:]):
            if len(idx) == 0:
                new_pred = self._outlier_response(neighbors)
            else:
                new_pred = self._local_regressor(self._y[idx],
                                                 distances[i + 1])

            pred = pred.concatenate(new_pred)

        return pred

    def _outlier_response(self, neighbors):
        """Response in case of no neighbors"""

        if (not hasattr(self, "outlier_response") or
                self.outlier_response is None):
            index = np.where([len(n) == 0 for n in neighbors])[0]

            raise ValueError(f"No neighbors found for test samples  {index}, "
                             "you can try using larger radius, give a reponse "
                             "for outliers, or consider removing them from "
                             "your dataset.")
        else:
            return self.outlier_response

    def score(self, X, y, sample_weight=None):
        r"""Return the coefficient of determination R^2 of the prediction.

        In the multivariate response case, the coefficient :math:`R^2` is
        defined as

        .. math::
            1 - \frac{\sum_{i=1}^{n} (y_i - \hat y_i)^2}
            {\sum_{i=1}^{n} (y_i - \frac{1}{n}\sum_{i=1}^{n}y_i)^2}

        where :math:`\hat{y}_i` is the prediction associated to the test sample
        :math:`X_i`, and :math:`{y}_i` is the true response. See
        :func:`sklearn.metrics.r2_score <sklearn.metrics.r2_score>` for more
        information.


        In the functional case it is returned an extension of the coefficient
        of determination :math:`R^2`, defined as

        .. math::
            1 - \frac{\sum_{i=1}^{n}\int (y_i(t) - \hat{y}_i(t))^2dt}
            {\sum_{i=1}^{n} \int (y_i(t)- \frac{1}{n}\sum_{i=1}^{n}y_i(t))^2dt}


        The best possible score is 1.0 and it can be negative
        (because the model can be arbitrarily worse). A constant model that
        always predicts the expected value of y, disregarding the input
        features, would get a R^2 score of 0.0.

        Args:
            X (FDataGrid): Test samples to be predicted.
            y (FData or array-like): True responses of the test samples.
            sample_weight (array_like, shape = [n_samples], optional): Sample
                weights.

        Returns:
            (float): Coefficient of determination.

        """
        if self._functional:
            return self._functional_score(X, y, sample_weight=sample_weight)
        else:
            # Default sklearn multivariate score
            return super().score(X, y, sample_weight=sample_weight)

    def _functional_score(self, X, y, sample_weight=None):
        r"""Return an extension of the coefficient of determination R^2.

        The coefficient is defined as

        .. math::
            1 - \frac{\sum_{i=1}^{n}\int (y_i(t) - \hat{y}_i(t))^2dt}
            {\sum_{i=1}^{n} \int (y_i(t)- \frac{1}{n}\sum_{i=1}^{n}y_i(t))^2dt}

        where :math:`\hat{y}_i` is the prediction associated to the test sample
        :math:`X_i`, and :math:`{y}_i` is the true response.

        The best possible score is 1.0 and it can be negative
        (because the model can be arbitrarily worse). A constant model that
        always predicts the expected value of y, disregarding the input
        features, would get a R^2 score of 0.0.

        Args:
            X (FDataGrid): Test samples to be predicted.
            y (FData): True responses of the test samples.
            sample_weight (array_like, shape = [n_samples], optional): Sample
                weights.

        Returns:
            (float): Coefficient of determination.

        """

        # TODO: If it is created a module in ml.regression with other
        # score metrics, move it.
        from scipy.integrate import simps

        if y.dim_codomain != 1 or y.dim_domain != 1:
            raise ValueError("Score not implemented for multivariate "
                             "functional data.")

        # Make prediction
        pred = self.predict(X)

        u = y - pred
        v = y - y.mean()

        # Discretize to integrate and make squares if needed
        if type(u) != FDataGrid:
            u = u.to_grid()
            v = v.to_grid()

        data_u = u.data_matrix[..., 0]
        data_v = v.data_matrix[..., 0]

        # Square without allocate more memory
        np.square(data_u, out=data_u)
        np.square(data_v, out=data_v)

        if sample_weight is not None:
            if len(sample_weight) != len(y):
                raise ValueError("Must be a weight for each sample.")

            sample_weight = np.asarray(sample_weight)
            sample_weight = sample_weight / sample_weight.sum()
            data_u_t = data_u.T
            data_u_t *= sample_weight
            data_v_t = data_v.T
            data_v_t *= sample_weight

        # Sum and integrate
        sum_u = np.sum(data_u, axis=0)
        sum_v = np.sum(data_v, axis=0)

        int_u = simps(sum_u, x=u.grid_points[0])
        int_v = simps(sum_v, x=v.grid_points[0])

        return 1 - int_u / int_v



import unittest

import numpy as np
from sklearn.exceptions import NotFittedError

from skfda import FDataGrid
from skfda._utils import invert_warping, normalize_warping
from skfda.datasets import (
    make_multimodal_landmarks,
    make_multimodal_samples,
    make_sinusoidal_process,
)
from skfda.exploratory.stats import mean
from skfda.preprocessing.registration import (
    LeastSquaresShiftRegistration,
    landmark_elastic_registration,
    landmark_elastic_registration_warping,
    landmark_shift_deltas,
    landmark_shift_registration,
)
from skfda.preprocessing.registration.validation import (
    AmplitudePhaseDecomposition,
    LeastSquares,
    PairwiseCorrelation,
    SobolevLeastSquares,
)
from skfda.representation.basis import Fourier
from skfda.representation.interpolation import SplineInterpolation


class TestWarping(unittest.TestCase):
    """Test warpings functions."""

    def setUp(self) -> None:
        """Initialize samples."""
        self.time = np.linspace(-1, 1, 50)
        interpolation = SplineInterpolation(3, monotone=True)
        self.polynomial = FDataGrid(
            [self.time**3, self.time**5],
            self.time,
            interpolation=interpolation,
        )

    def test_invert_warping(self) -> None:
        """Test that the composition with invert warping is the identity."""
        inverse = invert_warping(self.polynomial)

        # Check if identity
        identity = self.polynomial.compose(inverse)

        np.testing.assert_array_almost_equal(
            [self.time, self.time],
            identity.data_matrix[..., 0],
            decimal=3,
        )

    def test_standard_normalize_warping(self) -> None:
        """Test normalization to (0, 1)."""
        normalized = normalize_warping(self.polynomial, (0, 1))

        # Test new domain range (0, 1)
        np.testing.assert_array_equal(normalized.domain_range, [(0, 1)])

        np.testing.assert_array_almost_equal(
            normalized.grid_points[0],
            np.linspace(0, 1, 50),
        )

        np.testing.assert_array_almost_equal(
            normalized(0)[..., 0],
            [[0], [0]],
        )

        np.testing.assert_array_almost_equal(
            normalized(1)[..., 0],
            [[1.0], [1.0]],
        )

    def test_standard_normalize_warping_default_value(self) -> None:
        """Test normalization."""
        normalized = normalize_warping(self.polynomial)

        # Test new domain range (0, 1)
        np.testing.assert_array_equal(normalized.domain_range, [(-1, 1)])

        np.testing.assert_array_almost_equal(
            normalized.grid_points[0],
            np.linspace(-1, 1, 50),
        )

        np.testing.assert_array_almost_equal(
            normalized(-1)[..., 0],
            [[-1], [-1]],
        )

        np.testing.assert_array_almost_equal(
            normalized(1)[..., 0],
            [[1.0], [1.0]],
        )

    def test_normalize_warping(self) -> None:
        """Test normalization to (a, b)."""
        a = -4
        b = 3
        domain = (a, b)
        normalized = normalize_warping(self.polynomial, domain)

        # Test new domain range (0, 1)
        np.testing.assert_array_equal(normalized.domain_range, [domain])

        np.testing.assert_array_almost_equal(
            normalized.grid_points[0],
            np.linspace(*domain, 50),
        )

        np.testing.assert_array_equal(normalized(a)[..., 0], [[a], [a]])

        np.testing.assert_array_equal(normalized(b)[..., 0], [[b], [b]])

    def test_landmark_shift_deltas(self) -> None:
        """Test landmark shift deltas."""
        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()

        shifts = landmark_shift_deltas(fd, landmarks).round(3)
        np.testing.assert_almost_equal(shifts, [0.327, -0.173, -0.154])

    def test_landmark_shift_registration(self) -> None:
        """Test landmark shift registration."""
        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()

        original_modes = fd(
            landmarks.reshape((3, 1, 1)),
            aligned=False,
        )
        # Test default location
        fd_registered = landmark_shift_registration(fd, landmarks)
        center = np.mean(landmarks)
        reg_modes = fd_registered(center)

        # Test callable location
        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        fd_registered = landmark_shift_registration(
            fd,
            landmarks,
            location=np.mean,
        )
        center = np.mean(landmarks)
        reg_modes = fd_registered(center)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        # Test integer location
        fd_registered = landmark_shift_registration(
            fd,
            landmarks,
            location=0,
        )
        center = np.mean(landmarks)
        reg_modes = fd_registered(0)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

        # Test array location
        fd_registered = landmark_shift_registration(
            fd,
            landmarks,
            location=[0, 0.1, 0.2],
        )
        reg_modes = fd_registered([[0], [0.1], [0.2]], aligned=False)

        np.testing.assert_almost_equal(reg_modes, original_modes, decimal=2)

    def test_landmark_elastic_registration_warping(self) -> None:
        """Test the warpings in landmark elastic registration."""
        fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        landmarks = make_multimodal_landmarks(
            n_samples=3,
            n_modes=2,
            random_state=9,
        )
        landmarks = landmarks.squeeze()

        # Default location
        warping = landmark_elastic_registration_warping(fd, landmarks)
        center = (landmarks.max(axis=0) + landmarks.min(axis=0)) / 2
        np.testing.assert_almost_equal(
            warping(center)[..., 0],
            landmarks,
            decimal=1,
        )

        # Fixed location
        center = [0.3, 0.6]
        warping = landmark_elastic_registration_warping(
            fd, landmarks, location=center)
        np.testing.assert_almost_equal(
            warping(center)[..., 0],
            landmarks,
            decimal=3,
        )

    def test_landmark_elastic_registration(self) -> None:
        """Test landmark elastic registration."""
        fd = make_multimodal_samples(n_samples=3, n_modes=2, random_state=9)
        landmarks = make_multimodal_landmarks(
            n_samples=3,
            n_modes=2,
            random_state=9,
        )
        landmarks = landmarks.squeeze()

        original_values = fd(landmarks.reshape(3, 2), aligned=False)

        # Default location
        fd_reg = landmark_elastic_registration(fd, landmarks)
        center = (landmarks.max(axis=0) + landmarks.min(axis=0)) / 2
        np.testing.assert_almost_equal(
            fd_reg(center),
            original_values,
            decimal=2,
        )

        # Fixed location
        center = [.3, .6]
        fd_reg = landmark_elastic_registration(fd, landmarks, location=center)
        np.testing.assert_array_almost_equal(
            fd_reg(center),
            original_values,
            decimal=2,
        )


class TestLeastSquaresShiftRegistration(unittest.TestCase):
    """Test shift registration."""

    def setUp(self) -> None:
        """Initialize samples."""
        self.fd = make_sinusoidal_process(
            n_samples=2,
            error_std=0,
            random_state=1,
        )
        self.fd.extrapolation = "periodic"

    def test_fit_transform(self) -> None:

        reg = LeastSquaresShiftRegistration()

        # Test fit transform with FDataGrid
        fd_reg = reg.fit_transform(self.fd)

        # Check attributes fitted
        self.assertTrue(hasattr(reg, 'deltas_'))
        self.assertTrue(hasattr(reg, 'template_'))
        self.assertTrue(hasattr(reg, 'n_iter_'))
        self.assertTrue(isinstance(fd_reg, FDataGrid))

        deltas = reg.deltas_.round(3)
        np.testing.assert_array_almost_equal(deltas, [-0.022, 0.03])

        # Test with Basis
        fd = self.fd.to_basis(Fourier())
        reg.fit_transform(fd)
        deltas = reg.deltas_.round(3)
        np.testing.assert_array_almost_equal(deltas, [-0.022, 0.03])

    def test_fit_and_transform(self) -> None:
        """Test wrapper of shift_registration_deltas."""
        fd = make_sinusoidal_process(
            n_samples=2,
            error_std=0,
            random_state=10,
        )

        reg = LeastSquaresShiftRegistration()
        response = reg.fit(self.fd)

        # Check attributes and returned value
        self.assertTrue(hasattr(reg, 'template_'))
        self.assertTrue(response is reg)

        reg.transform(fd)
        deltas = reg.deltas_.round(3)
        np.testing.assert_allclose(deltas, [0.071, -0.072])

    def test_inverse_transform(self) -> None:

        reg = LeastSquaresShiftRegistration()
        fd = reg.fit_transform(self.fd)
        fd = reg.inverse_transform(fd)

        np.testing.assert_array_almost_equal(
            fd.data_matrix,
            self.fd.data_matrix,
            decimal=3,
        )

    def test_raises(self) -> None:

        reg = LeastSquaresShiftRegistration()

        # Test not fitted
        with np.testing.assert_raises(NotFittedError):
            reg.transform(self.fd)

        reg.fit(self.fd)
        reg.set_params(restrict_domain=True)

        # Test use transform with restrict_domain=True
        with np.testing.assert_raises(AttributeError):
            reg.transform(self.fd)

        # Test inverse_transform without previous transformation
        with np.testing.assert_raises(AttributeError):
            reg.inverse_transform(self.fd)

        reg.fit_transform(self.fd)

        # Test inverse transform with different number of sample
        with np.testing.assert_raises(ValueError):
            reg.inverse_transform(self.fd[:1])

        fd = make_multimodal_samples(dim_domain=2, random_state=0)

        with np.testing.assert_raises(ValueError):
            reg.fit_transform(fd)

        reg.set_params(initial=[0.])

        # Wrong initial estimation
        with np.testing.assert_raises(ValueError):
            reg.fit_transform(self.fd)

    def test_template(self) -> None:

        reg = LeastSquaresShiftRegistration()
        fd_registered_1 = reg.fit_transform(self.fd)

        reg_2 = LeastSquaresShiftRegistration(template=reg.template_)
        fd_registered_2 = reg_2.fit_transform(self.fd)

        reg_3 = LeastSquaresShiftRegistration(template=mean)
        fd_registered_3 = reg_3.fit_transform(self.fd)

        reg_4 = LeastSquaresShiftRegistration(template=reg.template_)
        fd_registered_4 = reg_4.fit(self.fd).transform(self.fd)

        np.testing.assert_array_almost_equal(
            fd_registered_1.data_matrix,
            fd_registered_3.data_matrix,
        )

        # With the template fixed could vary the convergence
        np.testing.assert_array_almost_equal(
            fd_registered_1.data_matrix,
            fd_registered_2.data_matrix,
            decimal=3,
        )

        np.testing.assert_array_almost_equal(
            fd_registered_2.data_matrix,
            fd_registered_4.data_matrix,
        )

    def test_restrict_domain(self) -> None:
        reg = LeastSquaresShiftRegistration(restrict_domain=True)
        fd_registered_1 = reg.fit_transform(self.fd)

        np.testing.assert_array_almost_equal(
            np.array(fd_registered_1.domain_range).round(3),
            [[0.022, 0.969]],
        )

        reg2 = LeastSquaresShiftRegistration(
            restrict_domain=True,
            template=reg.template_.copy(domain_range=self.fd.domain_range),
        )
        fd_registered_2 = reg2.fit_transform(self.fd)

        np.testing.assert_array_almost_equal(
            fd_registered_2.data_matrix,
            fd_registered_1.data_matrix,
            decimal=3,
        )

        reg3 = LeastSquaresShiftRegistration(
            restrict_domain=True,
            template=mean,
        )
        fd_registered_3 = reg3.fit_transform(self.fd)

        np.testing.assert_array_almost_equal(
            fd_registered_3.data_matrix,
            fd_registered_1.data_matrix,
        )

    def test_initial_estimation(self) -> None:
        reg = LeastSquaresShiftRegistration(initial=[-0.02161235, 0.03032652])
        reg.fit_transform(self.fd)

        # Only needed 1 iteration until convergence
        self.assertEqual(reg.n_iter_, 1)

    def test_custom_grid_points(self) -> None:
        reg = LeastSquaresShiftRegistration(grid_points=np.linspace(0, 1, 50))
        reg.fit_transform(self.fd)


class TestRegistrationValidation(unittest.TestCase):
    """Test validation functions."""

    def setUp(self) -> None:
        """Initialize the samples."""
        self.X = make_sinusoidal_process(error_std=0, random_state=0)
        self.shift_registration = LeastSquaresShiftRegistration().fit(self.X)

    def test_amplitude_phase_score(self) -> None:
        """Test basic usage of AmplitudePhaseDecomposition."""
        scorer = AmplitudePhaseDecomposition()
        score = scorer(self.shift_registration, self.X)
        np.testing.assert_allclose(score, 0.971144, rtol=1e-6)

    def test_amplitude_phase_score_with_basis(self) -> None:
        """Test the AmplitudePhaseDecomposition with FDataBasis."""
        scorer = AmplitudePhaseDecomposition()
        X = self.X.to_basis(Fourier())
        score = scorer(self.shift_registration, X)
        np.testing.assert_allclose(score, 0.995086, rtol=1e-6)

    def test_default_score(self) -> None:
        """Test default score of a registration transformer."""
        score = self.shift_registration.score(self.X)
        np.testing.assert_allclose(score, 0.971144, rtol=1e-6)

    def test_least_squares_score(self) -> None:
        """Test LeastSquares."""
        scorer = LeastSquares()
        score = scorer(self.shift_registration, self.X)
        np.testing.assert_allclose(score, 0.953355, rtol=1e-6)

    def test_sobolev_least_squares_score(self) -> None:
        """Test SobolevLeastSquares."""
        scorer = SobolevLeastSquares()
        score = scorer(self.shift_registration, self.X)
        np.testing.assert_allclose(score, 0.923962, rtol=1e-6)

    def test_pairwise_correlation(self) -> None:
        """Test PairwiseCorrelation."""
        scorer = PairwiseCorrelation()
        score = scorer(self.shift_registration, self.X)
        np.testing.assert_allclose(score, 1.816228, rtol=1e-6)

    def test_mse_decomposition(self) -> None:
        """Test obtaining all stats from AmplitudePhaseDecomposition."""
        fd = make_multimodal_samples(n_samples=3, random_state=1)
        landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        landmarks = landmarks.squeeze()
        warping = landmark_elastic_registration_warping(fd, landmarks)
        fd_registered = fd.compose(warping)
        scorer = AmplitudePhaseDecomposition()
        ret = scorer.stats(fd, fd_registered)
        np.testing.assert_allclose(ret.mse_amplitude, 0.0009465483)
        np.testing.assert_allclose(ret.mse_phase, 0.1051769136)
        np.testing.assert_allclose(ret.r_squared, 0.9910806875)
        np.testing.assert_allclose(ret.c_r, 0.9593073773)

    def test_raises_amplitude_phase(self) -> None:
        scorer = AmplitudePhaseDecomposition()

        # Inconsistent number of functions registered
        with np.testing.assert_raises(ValueError):
            scorer.score_function(self.X, self.X[:2])

        # Inconsistent number of functions registered
        with np.testing.assert_raises(ValueError):
            scorer.score_function(self.X, self.X[:-1])


if __name__ == '__main__':
    unittest.main()



"""Methods and classes for validation of the registration procedures."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

import numpy as np

from ..._utils import _to_grid, check_is_univariate
from ...representation import FData
from .base import RegistrationTransformer


class RegistrationScorer(ABC):
    """Cross validation scoring for registration procedures.

    It calculates the score of a registration procedure, used to perform
    model validation or parameter selection.

    Attributes:
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation and
            perform the calculation.

    Args:
        estimator (Estimator): Registration method estimator. The estimator
            should be fitted.
        X (:class:`FData <skfda.FData>`): Functional data to be registered.
        y (:class:`FData <skfda.FData>`, optional): Functional data target.
            If provided should be the same as `X` in general.

    Returns:
        float: Cross validation score.

    Note:
        The scorer passes the warpings generated in the registration procedure
        to the `score_function` when necessary.

    See also:
        :class:`~AmplitudePhaseDecomposition`
        :class:`~LeastSquares`
        :class:`~SobolevLeastSquares`
        :class:`~PairwiseCorrelation`

    """

    def __call__(
        self,
        estimator: RegistrationTransformer,
        X: FData,
        y: Optional[FData] = None,
    ) -> float:
        """Compute the score of the transformation.

        Args:
            estimator: Registration method estimator. The estimator
                should be fitted.
            X: Functional data to be registered.
            y: Functional data target. If provided should be the same as
                `X` in general.

        Returns:
            Cross validation score.
        """
        if y is None:
            y = X

        # Register the data
        X_reg = estimator.transform(X)

        return self.score_function(y, X_reg)

    @abstractmethod
    def score_function(
        self,
        X: FData,
        y: FData,
    ) -> float:
        """Compute the score of the transformation performed.

        Args:
            X: Original functional data.
            y: Functional data registered.

        Returns:
            Score of the transformation.

        """
        pass


@dataclass
class AmplitudePhaseDecompositionStats():
    r"""Named tuple to store the values of the amplitude-phase decomposition.

    Values of the amplitude phase decomposition computed in
    :func:`mse_r_squared`, returned when `return_stats` is `True`.

    Args:
        r_square (float): Squared correlation index :math:`R^2`.
        mse_amplitude (float): Mean square error of amplitude
            :math:`\text{MSE}_{amp}`.
        mse_phase (float): Mean square error of phase :math:`\text{MSE}_{pha}`.
        c_r (float): Constant :math:`C_R`.

    """

    r_squared: float
    mse_amplitude: float
    mse_phase: float
    c_r: float


class AmplitudePhaseDecomposition(
    RegistrationScorer,
):
    r"""Compute mean square error measures for amplitude and phase variation.

    Once the registration has taken place, this function computes two mean
    squared error measures, one for amplitude variation, and the other for
    phase variation and returns a squared multiple correlation index
    of the amount of variation in the unregistered functions is due to phase.

    Let :math:`x_i(t),y_i(t)` be the unregistered and registered functions
    respectively. The total mean square error measure (see [RGS09-8-5]_) is
    defined as


    .. math::
        \text{MSE}_{total}=
        \frac{1}{N}\sum_{i=1}^{N}\int[x_i(t)-\overline x(t)]^2dt

    The measures of amplitude and phase mean square error are

    .. math::
        \text{MSE}_{amp} =  C_R \frac{1}{N}
        \sum_{i=1}^{N} \int \left [ y_i(t) - \overline{y}(t) \right ]^2 dt

    .. math::
        \text{MSE}_{phase}=
        C_R \int \overline{y}^2(t) dt - \int \overline{x}^2(t) dt

    where the constant :math:`C_R` is defined as

    .. math::

        C_R = \frac{\frac{1}{N}\sum_{i=1}^{N}\int[x_i(t)-\overline x(t)]^2dt
        }{\frac{1}{N}\sum_{i=1}^{N}\int[y_i(t)-\overline y(t)]^2dt}

    whose structure is related to the covariation between the deformation
    functions :math:`Dh_i(t)` and the squared registered functions
    :math:`y_i^2(t)`. When these two sets of functions are independents
    :math:`C_R=1`, as in the case of shift registration.

    The total mean square error is decomposed in the two sources of
    variability.

    .. math::
        \text{MSE}_{total} = \text{MSE}_{amp} + \text{MSE}_{phase}

    The squared multiple correlation index of the proportion of the total
    variation due to phase is defined as:

    .. math::
        R^2 = \frac{\text{MSE}_{phase}}{\text{MSE}_{total}}

    See [KR08-3]_ for a detailed explanation.

    Attributes:
        return_stats (boolean, optional): If `true` returns a named tuple
            with four values: :math:`R^2`, :math:`MSE_{amp}`, :math:`MSE_{pha}`
            and :math:`C_R`. Otherwise the squared correlation index
            :math:`R^2` is returned. Default `False`.

        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation and
            perform the calculation.


    Args:
        estimator (RegistrationTransformer): Registration transformer.
        X (:class:`FData`): Unregistered functions.
        y (:class:`FData`, optional): Target data, generally the same as X. By
            default 'None', which uses `X` as target.


    Returns:
        (float or :class:`NamedTuple <typing.NamedTuple>`): squared correlation
        index :math:`R^2` if `return_stats` is `False`. Otherwise a named
        tuple containing:

            * `r_squared`: Squared correlation index :math:`R^2`.
            * `mse_amp`: Mean square error of amplitude
              :math:`\text{MSE}_{amp}`.
            * `mse_pha`: Mean square error of phase :math:`\text{MSE}_{pha}`.
            * `c_r`: Constant :math:`C_R`.


    Raises:
        ValueError: If the functional data is not univariate.

    References:
        ..  [KR08-3] Kneip, Alois & Ramsay, James. (2008).  Quantifying
            amplitude and phase variation. In *Combining Registration and
            Fitting for Functional Models* (pp. 14-15). Journal of the American
            Statistical Association.
        ..  [RGS09-8-5] Ramsay J.O., Giles Hooker & Spencer Graves (2009). In
            *Functional Data Analysis with R and Matlab* (pp. 125-126).
            Springer.

    Examples:
        Calculate the score of the shift registration of a sinusoidal process
        synthetically generated.

        >>> from skfda.preprocessing.registration.validation import \
        ...                                         AmplitudePhaseDecomposition
        >>> from skfda.preprocessing.registration import (
        ...     LeastSquaresShiftRegistration,
        ... )
        >>> from skfda.datasets import make_sinusoidal_process
        >>> X = make_sinusoidal_process(error_std=0, random_state=0)

        Fit the registration procedure.

        >>> shift_registration = LeastSquaresShiftRegistration()
        >>> shift_registration.fit(X)
        LeastSquaresShiftRegistration(...)

        Compute the :math:`R^2` correlation index

        >>> scorer = AmplitudePhaseDecomposition()
        >>> score = scorer(shift_registration, X)
        >>> round(score, 3)
        0.971

        Also it is possible to get all the values of the decomposition:

        >>> X_reg = shift_registration.transform(X)
        >>> stats = scorer.stats(X, X_reg)
        >>> round(stats.r_squared, 3)
        0.971
        >>> round(stats.mse_amplitude, 3)
        0.006
        >>> round(stats.mse_phase, 3)
        0.214
        >>> round(stats.c_r, 3)
        0.976


    See also:
        :class:`~LeastSquares`
        :class:`~SobolevLeastSquares`
        :class:`~PairwiseCorrelation`

    """

    def stats(
        self,
        X: FData,
        y: FData,
    ) -> AmplitudePhaseDecompositionStats:
        """
        Compute the decomposition statistics.

        Args:
            X: Original functional data.
            y: Functional data registered.

        Returns:
            The decomposition statistics.
        """
        from ...misc.metrics import l2_distance, l2_norm

        check_is_univariate(X)
        check_is_univariate(y)

        if len(y) != len(X):
            raise ValueError(
                f"The registered and unregistered curves must have "
                f"the same number of samples ({len(y)})!=({len(X)})",
            )

        X_mean = X.mean()
        y_mean = y.mean()

        c_r = np.sum(l2_norm(X)**2) / np.sum(l2_norm(y)**2)

        mse_amplitude = c_r * np.mean(l2_distance(y, y.mean())**2)
        mse_phase = (c_r * l2_norm(y_mean)**2 - l2_norm(X_mean)**2).item()

        # Should be equal to np.mean(l2_distance(X, X_mean)**2)
        mse_total = mse_amplitude + mse_phase

        # squared correlation measure of proportion of phase variation
        rsq = mse_phase / mse_total

        return AmplitudePhaseDecompositionStats(
            r_squared=rsq,
            mse_amplitude=mse_amplitude,
            mse_phase=mse_phase,
            c_r=c_r,
        )

    def score_function(
        self,
        X: FData,
        y: FData,
    ) -> float:
        """Compute the score of the transformation performed.

        Args:
            X: Original functional data.
            y: Functional data registered.

        Returns:
            Score of the transformation.

        """
        return float(self.stats(X, y).r_squared)


class LeastSquares(RegistrationScorer):
    r"""Cross-validated measure of the registration procedure.

    Computes a cross-validated measure of the level of synchronization
    [James07]_:

    .. math::
        ls=1 - \frac{1}{N} \sum_{i=1}^{N} \frac{\int\left(\tilde{f}_{i}(t)-
        \frac{1}{N-1} \sum_{j \neq i} \tilde{f}_{j}(t)\right)^{2} dt}{\int
        \left(f_{i}(t)-\frac{1}{N-1} \sum_{j \neq i} f_{j}(t)\right)^{2} dt}

    where :math:`f_i` and :math:`\tilde f_i` are the original and the
    registered data respectively.

    The :math:`ls` measures the total cross-sectional variance of the aligned
    functions, relative to the original value.
    A value of :math:`1` would indicate an identical shape for all registered
    curves, while zero corresponds to no improvement in the synchronization. It
    can be negative because the model can be arbitrarily worse.

    Attributes:
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation and
            perform the calculation.

    Args:
        estimator (RegistrationTransformer): Registration transformer.
        X (:class:`FData <skfda.FData>`): Original functional data.
        y (:class:`FData <skfda.FData>`): Registered functional data.


    Note:
        The original least square measure used in [S11-5-2-1]_ is defined as
        :math:`1 - ls`, but has been modified according to the scikit-learn
        scorers, where higher values correspond to better cross-validated
        measures.


    References:
        .. [James07] G. James. Curve alignments by moments. Annals of Applied
            Statistics, 1(2):480–501, 2007.
        .. [S11-5-2-1] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Comparisons with other Methods*
            (p. 18). arXiv:1103.3817v2.

    Examples:
        Calculate the score of the shift registration of a sinusoidal process
        synthetically generated.

        >>> from skfda.preprocessing.registration.validation import \
        ...                                                        LeastSquares
        >>> from skfda.preprocessing.registration import (
        ...     LeastSquaresShiftRegistration,
        ... )
        >>> from skfda.datasets import make_sinusoidal_process
        >>> X = make_sinusoidal_process(error_std=0, random_state=0)

        Fit the registration procedure.

        >>> shift_registration = LeastSquaresShiftRegistration()
        >>> shift_registration.fit(X)
        LeastSquaresShiftRegistration(...)

        Compute the least squares score.
        >>> scorer = LeastSquares()
        >>> score = scorer(shift_registration, X)
        >>> round(score, 3)
        0.953


    See also:
        :class:`~AmplitudePhaseDecomposition`
        :class:`~SobolevLeastSquares`
        :class:`~PairwiseCorrelation`

    """

    def score_function(self, X: FData, y: FData) -> float:
        """Compute the score of the transformation performed.

        Args:
            X (FData): Original functional data.
            y (FData): Functional data registered.

        Returns:
            float: Score of the transformation.

        """
        from ...misc.metrics import l2_distance

        check_is_univariate(X)
        check_is_univariate(y)

        # Instead of compute f_i - 1/(N-1) sum(j!=i)f_j for each i = 1 ... N
        # It is used (1 + 1/(N-1))f_i - 1/(N-1) sum(j=1 ... N) f_j =
        # (1 + 1/(N-1))f_i - N/(N-1) mean(f) =
        # C1 * f_1 - C2 mean(f) for each i= 1 ... N
        N = len(X)
        C1 = 1 + 1 / (N - 1)
        C2 = N / (N - 1)

        X = C1 * X
        y = C1 * y
        mean_X = C2 * X.mean()
        mean_y = C2 * y.mean()

        # Compute distance to mean
        ls_x = l2_distance(X, mean_X)**2
        ls_y = l2_distance(y, mean_y)**2

        # Quotient of distance
        quotient = ls_y / ls_x

        return float(1 - np.mean(quotient))


class SobolevLeastSquares(RegistrationScorer):
    r"""Cross-validated measure of the registration procedure.

    Computes a cross-validated measure of the level of synchronization
    [S11-5-2-3]_:

    .. math::
        sls=1 - \frac{\sum_{i=1}^{N} \int\left(\dot{\tilde{f}}_{i}(t)-
        \frac{1}{N} \sum_{j=1}^{N} \dot{\tilde{f}}_{j}\right)^{2} dt}
        {\sum_{i=1}^{N} \int\left(\dot{f}_{i}(t)-\frac{1}{N} \sum_{j=1}^{N}
        \dot{f}_{j}\right)^{2} dt}

    where :math:`\dot{f}_i` and :math:`\dot{\tilde{f}}_i` are the derivatives
    of the original and the registered data respectively.

    This criterion measures the total cross-sectional variance of the
    derivatives of the aligned functions, relative to the original value.
    A value of :math:`1` would indicate an identical shape for all registered
    curves, while zero corresponds to no improvement in the registration. It
    can be negative because the model can be arbitrarily worse.

    Attributes:
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation and
            perform the calculation.

    Args:
        estimator (RegistrationTransformer): Registration transformer.
        X (:class:`FData <skfda.FData>`): Original functional data.
        y (:class:`FData <skfda.FData>`): Registered functional data.

    Note:
        The original sobolev least square measure used in [S11-5-2-3]_ is
        defined as :math:`1 - sls`, but has been modified according to the
        scikit-learn scorers, where higher values correspond to better
        cross-validated measures.


    References:
        .. [S11-5-2-3] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Comparisons with other Methods*
            (p. 18). arXiv:1103.3817v2.

    Examples:
        Calculate the score of the shift registration of a sinusoidal process
        synthetically generated.

        >>> from skfda.preprocessing.registration.validation import \
        ...                                                 SobolevLeastSquares
        >>> from skfda.preprocessing.registration import (
        ...     LeastSquaresShiftRegistration,
        ... )
        >>> from skfda.datasets import make_sinusoidal_process
        >>> X = make_sinusoidal_process(error_std=0, random_state=0)

        Fit the registration procedure.

        >>> shift_registration = LeastSquaresShiftRegistration()
        >>> shift_registration.fit(X)
        LeastSquaresShiftRegistration(...)

        Compute the sobolev least squares score.
        >>> scorer = SobolevLeastSquares()
        >>> score = scorer(shift_registration, X)
        >>> round(score, 3)
        0.924

    See also:
        :class:`~AmplitudePhaseDecomposition`
        :class:`~LeastSquares`
        :class:`~PairwiseCorrelation`

    """

    def score_function(self, X: FData, y: FData) -> float:
        """Compute the score of the transformation performed.

        Args:
            X (FData): Original functional data.
            y (FData): Functional data registered.

        Returns:
            float: Score of the transformation.

        """
        from ...misc.metrics import l2_distance

        check_is_univariate(X)
        check_is_univariate(y)

        # Compute derivative
        X = X.derivative()
        y = y.derivative()

        # L2 distance to mean
        sls_x = l2_distance(X, X.mean())**2
        sls_y = l2_distance(y, y.mean())**2

        return float(1 - sls_y.sum() / sls_x.sum())


class PairwiseCorrelation(RegistrationScorer):
    r"""Cross-validated measure of pairwise correlation between functions.

    Computes a cross-validated pairwise correlation between functions
    to compare registration methods [S11-5-2-2]_ :

    .. math::
        pc=\frac{\sum_{i \neq j} \operatorname{cc}\left(\tilde{f}_{i}(t),
        \tilde{f}_{j}(t)\right)}{\sum_{i \neq j}
        \operatorname{cc}\left(f_{i}(t), f_{j}(t)\right)}

    where :math:`f_i` and :math:`\tilde f_i` are the original and registered
    data respectively and :math:`cc(f, g)` is the pairwise Pearson’s
    correlation between functions.

    The larger the value of :math:`pc`, the better the alignment between
    functions in general.

    Attributes:
        eval_points (array_like, optional): Set of points where the
            functions are evaluated to obtain a discrete representation and
            perform the calculation.

    Args:
        estimator (RegistrationTransformer): Registration transformer.
        X (:class:`FData <skfda.FData>`): Original functional data.
        y (:class:`FData <skfda.FData>`): Registered functional data.

    Note:
        Pearson’s correlation between functions is calculated assuming
        the samples are equiespaciated.

    References:
        .. [S11-5-2-2] Srivastava, Anuj et. al. Registration of Functional Data
            Using Fisher-Rao Metric (2011). In *Comparisons with other Methods*
            (p. 18). arXiv:1103.3817v2.

    Examples:
        Calculate the score of the shift registration of a sinusoidal process
        synthetically generated.

        >>> from skfda.preprocessing.registration.validation import \
        ...                                                 PairwiseCorrelation
        >>> from skfda.preprocessing.registration import (
        ...     LeastSquaresShiftRegistration,
        ... )
        >>> from skfda.datasets import make_sinusoidal_process
        >>> X = make_sinusoidal_process(error_std=0, random_state=0)

        Fit the registration procedure.

        >>> shift_registration = LeastSquaresShiftRegistration()
        >>> shift_registration.fit(X)
        LeastSquaresShiftRegistration(...)

        Compute the pairwise correlation score.
        >>> scorer = PairwiseCorrelation()
        >>> score = scorer(shift_registration, X)
        >>> round(score, 3)
        1.816

    See also:
        :class:`~AmplitudePhaseDecomposition`
        :class:`~LeastSquares`
        :class:`~SobolevLeastSquares`

    """

    def __init__(self, eval_points: Optional[np.ndarray] = None) -> None:
        self.eval_points = eval_points

    def score_function(self, X: FData, y: FData) -> float:
        """Compute the score of the transformation performed.

        Args:
            X (FData): Original functional data.
            y (FData): Functional data registered.

        Returns:
            float: Score of the transformation.

        """
        check_is_univariate(X)
        check_is_univariate(y)

        # Discretize functional data if needed
        X, y = _to_grid(X, y, eval_points=self.eval_points)

        # Compute correlation matrices with zeros in diagonal
        # corrcoefs computes the correlation between vector, without weights
        # due to the sample points
        X_corr = np.corrcoef(X.data_matrix[..., 0])
        np.fill_diagonal(X_corr, 0)

        y_corr = np.corrcoef(y.data_matrix[..., 0])
        np.fill_diagonal(y_corr, 0)

        return float(y_corr.sum() / X_corr.sum())



from skfda.inference.hotelling import hotelling_t2, hotelling_test_ind
from skfda.representation import FDataGrid
from skfda.representation.basis import Fourier
import unittest

import pytest


class HotellingTests(unittest.TestCase):

    def test_hotelling_test_ind_args(self):
        fd1 = FDataGrid([[1, 1, 1]])
        with self.assertRaises(TypeError):
            hotelling_test_ind(fd1, [])
        with self.assertRaises(TypeError):
            hotelling_test_ind([], fd1)
        with self.assertRaises(TypeError):
            hotelling_test_ind(fd1.to_basis(Fourier(n_basis=3)), fd1)
        with self.assertRaises(TypeError):
            hotelling_test_ind(fd1, fd1.to_basis(Fourier(n_basis=3)))
        with self.assertRaises(ValueError):
            hotelling_test_ind(fd1, fd1, n_reps=0)

    def test_hotelling_t2_args(self):
        fd1 = FDataGrid([[1, 1, 1]])
        with self.assertRaises(TypeError):
            hotelling_t2(fd1, [])
        with self.assertRaises(TypeError):
            hotelling_t2([], fd1)
        with self.assertRaises(TypeError):
            hotelling_t2(fd1.to_basis(Fourier(n_basis=3)), fd1)
        with self.assertRaises(TypeError):
            hotelling_t2(fd1, fd1.to_basis(Fourier(n_basis=3)))

    def test_hotelling_t2(self):
        fd1 = FDataGrid([[1, 1, 1], [1, 1, 1]])
        fd2 = FDataGrid([[1, 1, 1], [2, 2, 2]])
        self.assertAlmostEqual(hotelling_t2(fd1, fd1), 0)
        self.assertAlmostEqual(hotelling_t2(fd1, fd2), 1)

        fd1 = fd1.to_basis(Fourier(n_basis=3))
        fd2 = fd2.to_basis(Fourier(n_basis=3))
        self.assertAlmostEqual(hotelling_t2(fd1, fd1), 0)
        self.assertAlmostEqual(hotelling_t2(fd1, fd2), 1)

    def test_hotelling_test(self):
        fd1 = FDataGrid([[1, 1, 1], [1, 1, 1]])
        fd2 = FDataGrid([[3, 3, 3], [2, 2, 2]])
        t2, pval, dist = hotelling_test_ind(fd1, fd2, return_dist=True,
                                            random_state=0)
        self.assertAlmostEqual(t2, 9)
        self.assertAlmostEqual(pval, 0)
        self.assertEqual(len(dist), 6)
        reps = 5
        t2, pval, dist = hotelling_test_ind(fd1, fd2, return_dist=True,
                                            n_reps=reps, random_state=1)
        self.assertEqual(len(dist), reps)


if __name__ == '__main__':
    print()
    unittest.main()



import unittest

import numpy as np
import sklearn

import skfda
import skfda.preprocessing.smoothing as smoothing
import skfda.preprocessing.smoothing.kernel_smoothers as kernel_smoothers
import skfda.preprocessing.smoothing.validation as validation
from skfda._utils import _check_estimator
from skfda.misc.operators import LinearDifferentialOperator
from skfda.misc.regularization import L2Regularization
from skfda.representation.basis import BSpline, Monomial
from skfda.representation.grid import FDataGrid


class TestSklearnEstimators(unittest.TestCase):

    def test_nadaraya_watson(self):
        _check_estimator(kernel_smoothers.NadarayaWatsonSmoother)

    def test_local_linear_regression(self):
        _check_estimator(kernel_smoothers.LocalLinearRegressionSmoother)

    def test_knn(self):
        _check_estimator(kernel_smoothers.KNeighborsSmoother)


class _LinearSmootherLeaveOneOutScorerAlternative():
    r"""Alternative implementation of the LinearSmootherLeaveOneOutScorer"""

    def __call__(self, estimator, X, y):
        estimator_clone = sklearn.base.clone(estimator)

        estimator_clone._cv = True
        y_est = estimator_clone.fit_transform(X)

        return -np.mean((y.data_matrix[..., 0] - y_est.data_matrix[..., 0])**2)


class TestLeaveOneOut(unittest.TestCase):

    def _test_generic(self, estimator_class):
        loo_scorer = validation.LinearSmootherLeaveOneOutScorer()
        loo_scorer_alt = _LinearSmootherLeaveOneOutScorerAlternative()
        x = np.linspace(-2, 2, 5)
        fd = skfda.FDataGrid(x ** 2, x)

        estimator = estimator_class()

        grid = validation.SmoothingParameterSearch(
            estimator, [2, 3],
            scoring=loo_scorer)
        grid.fit(fd)
        score = np.array(grid.cv_results_['mean_test_score'])

        grid_alt = validation.SmoothingParameterSearch(
            estimator, [2, 3],
            scoring=loo_scorer_alt)
        grid_alt.fit(fd)
        score_alt = np.array(grid_alt.cv_results_['mean_test_score'])

        np.testing.assert_array_almost_equal(score, score_alt)

    def test_nadaraya_watson(self):
        self._test_generic(kernel_smoothers.NadarayaWatsonSmoother)

    def test_local_linear_regression(self):
        self._test_generic(kernel_smoothers.LocalLinearRegressionSmoother)

    def test_knn(self):
        self._test_generic(kernel_smoothers.KNeighborsSmoother)


class TestBasisSmoother(unittest.TestCase):

    def test_cholesky(self):
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), n_basis=5)
        fd = FDataGrid(data_matrix=x, grid_points=t)
        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=10,
            regularization=L2Regularization(
                LinearDifferentialOperator(2)),
            method='cholesky',
            return_basis=True)
        fd_basis = smoother.fit_transform(fd)
        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.60, 0.47, 0.20, -0.07, -0.20]])
        )

    def test_qr(self):
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSpline((0, 1), n_basis=5)
        fd = FDataGrid(data_matrix=x, grid_points=t)
        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=10,
            regularization=L2Regularization(
                LinearDifferentialOperator(2)),
            method='qr',
            return_basis=True)
        fd_basis = smoother.fit_transform(fd)
        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.60, 0.47, 0.20, -0.07, -0.20]])
        )

    def test_monomial_smoothing(self):
        # It does not have much sense to apply smoothing in this basic case
        # where the fit is very good but its just for testing purposes
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = Monomial(n_basis=4)
        fd = FDataGrid(data_matrix=x, grid_points=t)
        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=1,
            regularization=L2Regularization(
                LinearDifferentialOperator(2)),
            return_basis=True)
        fd_basis = smoother.fit_transform(fd)
        # These results where extracted from the R package fda
        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.61, -0.88, 0.06, 0.02]]))

    def test_vector_valued_smoothing(self) -> None:
        X, _ = skfda.datasets.fetch_weather(return_X_y=True)

        basis_dim = skfda.representation.basis.Fourier(
            n_basis=7, domain_range=X.domain_range)
        basis = skfda.representation.basis.VectorValued(
            [basis_dim] * 2
        )

        for method in ('cholesky', 'qr', 'svd'):
            with self.subTest(method=method):

                basis_smoother = smoothing.BasisSmoother(
                    basis,
                    regularization=L2Regularization(
                        LinearDifferentialOperator(2)),
                    return_basis=True,
                    smoothing_parameter=1,
                    method=method)

                basis_smoother_dim = smoothing.BasisSmoother(
                    basis_dim,
                    regularization=L2Regularization(
                        LinearDifferentialOperator(2)),
                    return_basis=True,
                    smoothing_parameter=1,
                    method=method)

                X_basis = basis_smoother.fit_transform(X)

                self.assertEqual(X_basis.dim_codomain, 2)

                self.assertEqual(X_basis.coordinates[0].basis, basis_dim)
                np.testing.assert_allclose(
                    X_basis.coordinates[0].coefficients,
                    basis_smoother_dim.fit_transform(
                        X.coordinates[0]).coefficients)

                self.assertEqual(X_basis.coordinates[1].basis, basis_dim)
                np.testing.assert_allclose(
                    X_basis.coordinates[1].coefficients,
                    basis_smoother_dim.fit_transform(
                        X.coordinates[1]).coefficients)



from skfda.datasets import fetch_gait
from skfda.inference.anova import oneway_anova, v_asymptotic_stat, \
    v_sample_stat
from skfda.representation import FDataGrid
from skfda.representation.basis import Fourier
import unittest

import pytest

import numpy as np


class OnewayAnovaTests(unittest.TestCase):

    def test_oneway_anova_args(self):
        with self.assertRaises(ValueError):
            oneway_anova()
        with self.assertRaises(ValueError):
            oneway_anova(1, '2')
        with self.assertRaises(ValueError):
            oneway_anova(FDataGrid([0]), n_reps=-2)

    def test_v_stats_args(self):
        with self.assertRaises(ValueError):
            v_sample_stat(1, [1])
        with self.assertRaises(ValueError):
            v_sample_stat(FDataGrid([0]), [0, 1])
        with self.assertRaises(ValueError):
            v_asymptotic_stat(1, [1])
        with self.assertRaises(ValueError):
            v_asymptotic_stat(FDataGrid([0]), [0, 1])
        with self.assertRaises(ValueError):
            v_asymptotic_stat(FDataGrid([[1, 1, 1], [1, 1, 1]]), [0, 0])

    def test_v_stats(self):
        n_features = 50
        weights = [1, 2, 3]
        t = np.linspace(0, 1, n_features)
        m1 = [1 for _ in range(n_features)]
        m2 = [2 for _ in range(n_features)]
        m3 = [3 for _ in range(n_features)]
        fd = FDataGrid([m1, m2, m3], grid_points=t)
        self.assertEqual(v_sample_stat(fd, weights), 7.0)
        self.assertAlmostEqual(v_sample_stat(fd.to_basis(Fourier(n_basis=5)),
                                             weights), 7.0)
        res = (1 - 2 * np.sqrt(1 / 2)) ** 2 + (1 - 3 * np.sqrt(1 / 3)) ** 2 \
                                            + (2 - 3 * np.sqrt(2 / 3)) ** 2
        self.assertAlmostEqual(v_asymptotic_stat(fd, weights), res)
        self.assertAlmostEqual(v_asymptotic_stat(fd.to_basis(Fourier(
            n_basis=5)), weights), res)

    def test_asymptotic_behaviour(self):
        dataset = fetch_gait()
        fd = dataset['data'].coordinates[1]
        fd1 = fd[0:5]
        fd2 = fd[5:10]
        fd3 = fd[10:15]

        n_little_sim = 10

        sims = np.array([oneway_anova(
            fd1, fd2, fd3, n_reps=500, random_state=i)[1]
            for i in range(n_little_sim)])
        little_sim = np.mean(sims)
        big_sim = oneway_anova(fd1, fd2, fd3, n_reps=2000, random_state=100)[1]
        self.assertAlmostEqual(little_sim, big_sim, delta=0.05)

        fd = fd.to_basis(Fourier(n_basis=5))
        fd1 = fd[0:5]
        fd2 = fd[5:10]

        sims = np.array([oneway_anova(
            fd1, fd2, n_reps=500, random_state=i)[1]
            for i in range(n_little_sim)])
        little_sim = np.mean(sims)
        big_sim = oneway_anova(fd1, fd2, n_reps=2000, random_state=100)[1]
        self.assertAlmostEqual(little_sim, big_sim, delta=0.05)


if __name__ == '__main__':
    print()
    unittest.main()



"""Utilities for norms and metrics."""
from typing import Any, Callable, Generic, Optional, Tuple, TypeVar

import multimethod
import numpy as np

from ..._utils import _pairwise_symmetric
from ...representation import FData, FDataGrid
from ...representation._typing import NDArrayFloat, Vector
from ._typing import Metric, MetricElementType, Norm, VectorType

T = TypeVar("T", bound=FData)


def _check_compatible(fdata1: T, fdata2: T) -> None:

    if isinstance(fdata1, FData) and isinstance(fdata2, FData):
        if (
            fdata2.dim_codomain != fdata1.dim_codomain
            or fdata2.dim_domain != fdata1.dim_domain
        ):
            raise ValueError("Objects should have the same dimensions")

        if not np.array_equal(fdata1.domain_range, fdata2.domain_range):
            raise ValueError("Domain ranges for both objects must be equal")


def _cast_to_grid(
    fdata1: FData,
    fdata2: FData,
    eval_points: Optional[NDArrayFloat] = None,
    _check: bool = True,
) -> Tuple[FDataGrid, FDataGrid]:
    """
    Convert fdata1 and fdata2 to FDatagrid.

    Checks if the fdatas passed as argument are unidimensional and compatible
    and converts them to FDatagrid to compute their distances.

    Args:
        fdata1: First functional object.
        fdata2: Second functional object.
        eval_points: Evaluation points.

    Returns:
        Tuple with two :obj:`FDataGrid` with the same grid points.
    """
    # Dont perform any check
    if not _check:
        return fdata1, fdata2

    _check_compatible(fdata1, fdata2)

    # Case new evaluation points specified
    if eval_points is not None:  # noqa: WPS223
        fdata1 = fdata1.to_grid(eval_points)
        fdata2 = fdata2.to_grid(eval_points)

    elif not isinstance(fdata1, FDataGrid) and isinstance(fdata2, FDataGrid):
        fdata1 = fdata1.to_grid(fdata2.grid_points[0])

    elif not isinstance(fdata2, FDataGrid) and isinstance(fdata1, FDataGrid):
        fdata2 = fdata2.to_grid(fdata1.grid_points[0])

    elif (
        not isinstance(fdata1, FDataGrid)
        and not isinstance(fdata2, FDataGrid)
    ):
        domain = fdata1.domain_range[0]
        grid_points = np.linspace(*domain)
        fdata1 = fdata1.to_grid(grid_points)
        fdata2 = fdata2.to_grid(grid_points)

    elif not np.array_equal(
        fdata1.grid_points,
        fdata2.grid_points,
    ):
        raise ValueError(
            "Grid points for both objects must be equal or"
            "a new list evaluation points must be specified",
        )

    return fdata1, fdata2


class NormInducedMetric(Metric[VectorType]):
    r"""
    Metric induced by a norm.

    Given a norm :math:`\| \cdot \|: X \rightarrow \mathbb{R}`,
    returns the metric :math:`d: X \times X \rightarrow \mathbb{R}` induced
    by the norm:

    .. math::
        d(f,g) = \|f - g\|

    Args:
        norm: Norm used to induce the metric.

    Examples:
        Computes the :math:`\mathbb{L}^2` distance between an object containing
        functional data corresponding to the function :math:`y(x) = x` defined
        over the interval [0, 1] and another one containing data of the
        function :math:`y(x) = x/2`.

        Firstly we create the functional data.

        >>> import skfda
        >>> import numpy as np
        >>> from skfda.misc.metrics import l2_norm, NormInducedMetric
        >>>
        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([x], x)
        >>> fd2 = skfda.FDataGrid([x/2], x)

        To construct the :math:`\mathbb{L}^2` distance it is used the
        :math:`\mathbb{L}^2` norm wich it is used to compute the distance.

        >>> l2_distance = NormInducedMetric(l2_norm)
        >>> d = l2_distance(fd, fd2)
        >>> float('%.3f'% d)
        0.289

    """

    def __init__(self, norm: Norm[VectorType]):
        self.norm = norm

    def __call__(self, elem1: VectorType, elem2: VectorType) -> NDArrayFloat:
        """Compute the induced norm between two vectors."""
        return self.norm(elem1 - elem2)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(norm={self.norm})"


@multimethod.multidispatch
def pairwise_metric_optimization(
    metric: Any,
    elem1: Any,
    elem2: Optional[Any],
) -> NDArrayFloat:
    r"""
    Optimized computation of a pairwise metric.

    This is a generic function that can be subclassed for different
    combinations of metric and operators in order to provide a more
    efficient implementation for the pairwise metric matrix.
    """
    return NotImplemented


class PairwiseMetric(Generic[MetricElementType]):
    r"""Pairwise metric function.

    Computes a given metric pairwise. The matrix returned by the pairwise
    metric is a matrix with as many rows as observations in the first object
    and as many columns as observations in the second one. Each element
    (i, j) of the matrix is the distance between the ith observation of the
    first object and the jth observation of the second one.

    Args:
        metric: Metric between two elements of a metric
            space.

    """

    def __init__(
        self,
        metric: Metric[MetricElementType],
    ):
        self.metric = metric

    def __call__(
        self,
        elem1: MetricElementType,
        elem2: Optional[MetricElementType] = None,
    ) -> NDArrayFloat:
        """Evaluate the pairwise metric."""
        optimized = pairwise_metric_optimization(self.metric, elem1, elem2)

        return (
            _pairwise_symmetric(self.metric, elem1, elem2)
            if optimized is NotImplemented
            else optimized
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}(metric={self.metric})"


Original = TypeVar("Original", bound=Vector)
Transformed = TypeVar("Transformed", bound=Vector)


class TransformationMetric(Generic[Original, Transformed], Metric[Original]):
    """
    Compute a distance after transforming the data.

    This is a convenience function to compute a metric after a transformation
    is applied to the data. It can be used, for example, to compute
    Sobolev-like metrics.

    Args:
        e1: First object.
        e2: Second object.

    Returns:
        Distance.

    Examples:
        Compute the L2 distance between the function derivatives.

        >>> import skfda
        >>> from skfda.misc.metrics import l2_distance, TransformationMetric

        >>> x = np.linspace(0, 1, 1001)
        >>> fd = skfda.FDataGrid([x], x)
        >>> fd2 = skfda.FDataGrid([x/2], x)

        >>> dist = TransformationMetric(
        ...     transformation=lambda x: x.derivative(),
        ...     metric=l2_distance,
        ... )
        >>> dist(fd, fd2)
        array([ 0.5])

    """

    def __init__(
        self,
        transformation: Callable[[Original], Transformed],
        metric: Metric[Transformed],
    ):
        self.transformation = transformation
        self.metric = metric

    def __call__(
        self,
        e1: Original,
        e2: Original,
    ) -> NDArrayFloat:
        """Compute the distance."""
        e1_trans = self.transformation(e1)
        e2_trans = self.transformation(e2)

        return self.metric(e1_trans, e2_trans)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}()"
        )


@pairwise_metric_optimization.register
def _pairwise_metric_optimization_transformation_distance(
    metric: TransformationMetric[Any, Any],
    e1: T,
    e2: Optional[T],
) -> NDArrayFloat:

    e1_trans = metric.transformation(e1)
    e2_trans = None if e2 is None else metric.transformation(e2)

    pairwise = PairwiseMetric(metric.metric)

    return pairwise(e1_trans, e2_trans)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# scikit-fda documentation build configuration file, created by
# sphinx-quickstart on Sun Oct 22 18:46:59 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#

import os
import sys
import warnings

import pkg_resources
# -- Extensions to the  Napoleon GoogleDocstring class ---------------------
from sphinx.ext.napoleon.docstring import GoogleDocstring

try:
    release = pkg_resources.get_distribution('scikit-fda').version
except pkg_resources.DistributionNotFound:
    print('To build the documentation, The distribution information of\n'
          'scikit-fda has to be available.  Either install the package\n'
          'into your development environment or run "setup.py develop"\n'
          'to setup the metadata.  A virtualenv is recommended!\n')
    sys.exit(1)
del pkg_resources

version = '.'.join(release.split('.')[:2])

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.todo',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon',
              'sphinx.ext.mathjax',
              'sphinx_rtd_theme',
              'sphinx_gallery.gen_gallery',
              'sphinx.ext.intersphinx',
              'sphinx.ext.doctest',
              'jupyter_sphinx',
              'sphinx.ext.autodoc.typehints',
              'sphinxcontrib.bibtex']

bibtex_bibfiles = ['refs.bib']

autodoc_default_flags = ['members', 'inherited-members']


doctest_global_setup = '''
import numpy as np
np.set_printoptions(legacy='1.13')
'''

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'scikit-fda'
copyright = ('2019, Grupo de Aprendizaje Automático - ' +
             'Universidad Autónoma de Madrid')
author = 'Author'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_logo = "logos/notitle_logo/notitle_logo.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'logo_only': True,
    'style_nav_header_background': 'Gainsboro',
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
        'donate.html',
    ]
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'skfdadoc'

# -- Options for LaTeX output ---------------------------------------------

latex_engine = 'lualatex'

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'skfda.tex', 'scikit-fda Documentation',
     'Author', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'skfda', 'scikit-fda Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'skfda', 'scikit-fda Documentation',
     author, 'skfda', 'One line description of project.',
     'Miscellaneous'),
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']

intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(
        sys.version_info), None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'sklearn': ('https://scikit-learn.org/stable', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
}


tutorial_list = [
    "plot_introduction.py",
    "plot_getting_data.py",
    "plot_basis_representation.py",
    "plot_skfda_sklearn.py",
]


class SkfdaExplicitSubOrder(object):
    """
    Class for use within the 'within_subsection_order' key.

    Inspired by Matplotlib gallery.

    """

    def __init__(self, src_dir: str) -> None:
        self.src_dir = src_dir  # src_dir is unused here
        self.ordered_list = tutorial_list

    def __call__(self, filename: str) -> str:
        """Return a string determining the sort order."""
        if filename in self.ordered_list:
            ind = self.ordered_list.index(filename)
            return f"{ind:04d}"

        # ensure not explicitly listed items come last.
        return f"zzz{filename}"


sphinx_gallery_conf = {
    # path to your examples scripts
    'examples_dirs': ['../examples', '../tutorial'],
    # path where to save gallery generated examples
    'gallery_dirs': ['auto_examples', 'auto_tutorial'],
    'reference_url': {
        # The module you locally document uses None
        'skfda': None,
    },
    'backreferences_dir': 'backreferences',
    'doc_module': 'skfda',
    'within_subsection_order': SkfdaExplicitSubOrder,
}

warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message='Matplotlib is currently using agg, which is a'
    ' non-GUI backend, so cannot show the figure.',
)

autosummary_generate = True
autodoc_typehints = "description"
napoleon_use_rtype = True

autodoc_type_aliases = {
    "ArrayLike": "ArrayLike",
    "GridPointsLike": "Union[ArrayLike, Sequence[ArrayLike]]",
}

# Napoleon fix for attributes
# Taken from
# https://michaelgoerz.net/notes/extending-sphinx-napoleon-docstring-sections.html


# first, we define new methods for any new sections and add them to the class


def parse_keys_section(self, section):
    return self._format_fields('Keys', self._consume_fields())


GoogleDocstring._parse_keys_section = parse_keys_section


def parse_attributes_section(self, section):
    return self._format_fields('Attributes', self._consume_fields())


GoogleDocstring._parse_attributes_section = parse_attributes_section


def parse_class_attributes_section(self, section):
    return self._format_fields('Class Attributes', self._consume_fields())


GoogleDocstring._parse_class_attributes_section = parse_class_attributes_section

# we now patch the parse method to guarantee that the the above methods are
# assigned to the _section dict


def patched_parse(self):
    self._sections['keys'] = self._parse_keys_section
    self._sections['class attributes'] = self._parse_class_attributes_section
    self._unpatched_parse()


GoogleDocstring._unpatched_parse = GoogleDocstring._parse
GoogleDocstring._parse = patched_parse



from __future__ import annotations

import abc
from typing import Any, Callable, TypeVar, Union

import multimethod
import numpy as np
from typing_extensions import Protocol

from ...representation import FData
from ...representation.basis import Basis

OperatorInput = TypeVar(
    "OperatorInput",
    bound=Union[np.ndarray, FData, Basis],
    contravariant=True,
)

OutputType = Union[np.ndarray, Callable[[np.ndarray], np.ndarray]]

OperatorOutput = TypeVar(
    "OperatorOutput",
    bound=OutputType,
    covariant=True,
)


class Operator(Protocol[OperatorInput, OperatorOutput]):
    """Abstract class for :term:`operators`."""

    @abc.abstractmethod
    def __call__(self, vector: OperatorInput) -> OperatorOutput:
        """Evaluate the operator."""
        pass


@multimethod.multidispatch
def gramian_matrix_optimization(
    linear_operator: Any,
    basis: OperatorInput,
) -> np.ndarray:
    """
    Efficient implementation of gramian_matrix.

    Generic function that can be subclassed for different combinations of
    operator and basis in order to provide a more efficient implementation
    for the gramian matrix.
    """
    return NotImplemented


def gramian_matrix_numerical(
    linear_operator: Operator[OperatorInput, OutputType],
    basis: OperatorInput,
) -> np.ndarray:
    """
    Return the gramian matrix given a basis, computed numerically.

    This method should work for every linear operator.

    """
    from .. import inner_product_matrix

    evaluated_basis = linear_operator(basis)

    domain_range = getattr(basis, "domain_range", None)

    return inner_product_matrix(evaluated_basis, _domain_range=domain_range)


def gramian_matrix(
    linear_operator: Operator[OperatorInput, OutputType],
    basis: OperatorInput,
) -> np.ndarray:
    r"""
    Return the gramian matrix given a basis.

    The gramian operator of a linear operator :math:`\Gamma` is

    .. math::
        G = \Gamma*\Gamma

    This method evaluates that gramian operator in a given basis,
    which is necessary for performing Tikhonov regularization,
    among other things.

    It tries to use an optimized implementation if one is available,
    falling back to a numerical computation otherwise.

    """
    # Try to use a more efficient implementation
    matrix = gramian_matrix_optimization(linear_operator, basis)
    if matrix is not NotImplemented:
        return matrix

    return gramian_matrix_numerical(linear_operator, basis)


class MatrixOperator(Operator[np.ndarray, np.ndarray]):
    """Linear operator for finite spaces.

    Between finite dimensional spaces, every linear operator can be expressed
    as a product by a matrix.

    Attributes:
        matrix:  The matrix containing the linear transformation.

    """

    def __init__(self, matrix: np.ndarray) -> None:
        self.matrix = matrix

    def __call__(self, f: np.ndarray) -> np.ndarray:  # noqa: D102
        return self.matrix @ f



import itertools
from typing import Optional, Tuple, Union, overload

import numpy as np
from sklearn.utils import check_random_state
from typing_extensions import Literal

import scipy.special

from ..._utils import RandomStateLike
from ...representation import FData, FDataBasis


def hotelling_t2(
    fd1: FData,
    fd2: FData,
) -> float:
    r"""
    Compute Hotelling's :math:`T^2` statistic.

    Calculates Hotelling's :math:`T^2` over two samples in
    :class:`skfda.representation.FData` objects with sizes :math:`n_1`
    and :math:`n_2`.

    .. math::
        T^2 = n(\mathbf{m}_1 - \mathbf{m}_2)^\top \mathbf{W}^{1/2}(
        \mathbf{W}^{1/2}\mathbf{K_{\operatorname{pooled}}} \mathbf{W}^{
        1/2})^+
        \mathbf{W}^{1/2} (\mathbf{m}_1 - \mathbf{m}_2),

    where :math:`(\cdot)^{+}` indicates the Moore-Penrose pseudo-inverse
    operator, :math:`n=n_1+n_2`, `W` is Gram matrix (identity in case of
    discretized data), :math:`\mathbf{m}_1, \mathbf{m}_2` are the
    means of each ample and :math:`\mathbf{K}_{\operatorname{pooled}}`
    matrix is defined as

    .. math::
        \mathbf{K}_{\operatorname{pooled}} :=
        \cfrac{n_1 - 1}{n_1 + n_2 - 2} \mathbf{K}_{n_1} +
        \cfrac{n_2 - 1}{n_1 + n_2 - 2} \mathbf{K}_{n_2},

    where :math:`\mathbf{K}_{n_1}`, :math:`\mathbf{K}_{n_2}` are the sample
    covariance matrices, computed with the basis coefficients or using
    the discrete representation, depending on the input.

    This statistic is defined in Pini, Stamm and Vantini
    :footcite:`pini+stamm+vantini_2018_hotellings`.

    Args:
        fd1: Object with the first sample.
        fd2: Object containing second sample.

    Returns:
        The value of the statistic.

    Raises:
        TypeError.

    Examples:
        >>> from skfda.inference.hotelling import hotelling_t2
        >>> from skfda.representation import FDataGrid, basis

        >>> fd1 = FDataGrid([[1, 1, 1], [3, 3, 3]])
        >>> fd2 = FDataGrid([[3, 3, 3], [5, 5, 5]])
        >>> '%.2f' % hotelling_t2(fd1, fd2)
        '2.00'
        >>> fd1 = fd1.to_basis(basis.Fourier(n_basis=3))
        >>> fd2 = fd2.to_basis(basis.Fourier(n_basis=3))
        >>> '%.2f' % hotelling_t2(fd1, fd2)
        '2.00'

    References:
        .. footbibliography::

    """
    if not isinstance(fd1, FData):
        raise TypeError("Argument type must inherit FData.")

    if not isinstance(fd2, type(fd1)):
        raise TypeError("Both samples must be instances of the same type.")

    n1, n2 = fd1.n_samples, fd2.n_samples  # Size of each sample
    n = n1 + n2  # Size of full sample
    m = fd1.mean() - fd2.mean()  # Delta mean

    if isinstance(fd1, FDataBasis):
        if fd1.basis != fd2.basis:
            raise ValueError(
                "Both FDataBasis objects must share the same basis.",
            )
        # When working on basis representation we use the coefficients
        m = m.coefficients[0]
        k1 = np.cov(fd1.coefficients, rowvar=False)
        k2 = np.cov(fd2.coefficients, rowvar=False)
        # If no weight matrix is passed, then we compute the Gram Matrix
        weights = fd1.basis.gram_matrix()
        weights = np.sqrt(weights)
    else:
        # Working with standard discretized data
        m = m.data_matrix[0, ..., 0]
        k1 = fd1.cov().data_matrix[0, ..., 0]
        k2 = fd2.cov().data_matrix[0, ..., 0]

    m = m.reshape((-1, 1))  # Reshaping the mean for a proper matrix product
    k_pool = ((n1 - 1) * k1 + (n2 - 1) * k2) / (n - 2)  # Combination of covs

    if isinstance(fd1, FDataBasis):
        assert weights is not None

        # Product of pooled covariance with the weights and Moore-Penrose inv.
        k_inv = np.linalg.pinv(np.linalg.multi_dot([weights, k_pool, weights]))
        k_inv = weights.dot(k_inv).dot(weights)
    else:
        # If data is discrete no weights are needed
        k_inv = np.linalg.pinv(k_pool)

    return float(n1 * n2 / n * m.T.dot(k_inv).dot(m)[0][0])


@overload
def hotelling_test_ind(
    fd1: FData,
    fd2: FData,
    *,
    n_reps: Optional[int] = None,
    random_state: RandomStateLike = None,
    return_dist: Literal[False] = False,
) -> Tuple[float, float]:
    pass


@overload
def hotelling_test_ind(
    fd1: FData,
    fd2: FData,
    *,
    n_reps: Optional[int] = None,
    random_state: RandomStateLike = None,
    return_dist: Literal[True],
) -> Tuple[float, float, np.ndarray]:
    pass


def hotelling_test_ind(
    fd1: FData,
    fd2: FData,
    *,
    n_reps: Optional[int] = None,
    random_state: RandomStateLike = None,
    return_dist: bool = False,
) -> Union[Tuple[float, float], Tuple[float, float, np.ndarray]]:
    """
    Compute Hotelling :math:`T^2`-test.

    Calculate the :math:`T^2`-test for the means of two independent samples of
    functional data.

    This is a two-sided test for the null hypothesis that 2 independent samples
    have identical average (expected) values. This test assumes that the
    populations have identical variances by default.

    The p-value of the test is calculated using a permutation test over the
    statistic :func:`~skfda.inference.hotelling.hotelling_t2`. If a maximum
    number of repetitions of the algorithm is provided then the permutations
    tested are generated randomly.

    This procedure is from Pini, Stamm and Vantinni
    :footcite:`pini+stamm+vantini_2018_hotellings`.

    Args:
        fd1: First sample of data.
        fd2: Second sample of data. The data objects must have the same
            type.
        n_reps: Maximum number of repetitions to compute
            p-value. Default value is None.
        random_state: Random state.
        return_dist: Flag to indicate if the function should
            return a numpy.array with the values of the statistic computed over
            each permutation.

    Returns:
        Value of the sample statistic, one tailed p-value and a collection of
        statistic values from permutations of the sample.

    Raises:
        TypeError: In case of bad arguments.

    Examples:
        >>> from skfda.inference.hotelling import hotelling_t2
        >>> from skfda.representation import FDataGrid, basis
        >>> from numpy import printoptions

        >>> fd1 = FDataGrid([[1, 1, 1], [3, 3, 3]])
        >>> fd2 = FDataGrid([[3, 3, 3], [5, 5, 5]])
        >>> t2n, pval, dist = hotelling_test_ind(fd1, fd2, return_dist=True)
        >>> '%.2f' % t2n
        '2.00'
        >>> '%.2f' % pval
        '0.00'
        >>> with printoptions(precision=4):
        ...     print(dist)
        [ 2. 2. 0. 0. 2. 2.]

    References:
        .. footbibliography::

    """
    if not isinstance(fd1, FData):
        raise TypeError("Argument type must inherit FData.")

    if not isinstance(fd2, type(fd1)):
        raise TypeError("Both samples must be instances of the same type.")

    if n_reps is not None and n_reps < 1:
        raise ValueError("Number of repetitions must be positive.")

    n1, n2 = fd1.n_samples, fd2.n_samples
    t2_0 = hotelling_t2(fd1, fd2)
    n = n1 + n2
    sample = fd1.concatenate(fd2)
    indices = np.arange(n)

    if n_reps is not None:  # Computing n_reps random permutations
        random_state = check_random_state(random_state)
        dist = np.empty(n_reps)
        for i in range(n_reps):
            random_state.shuffle(indices)
            dist[i] = hotelling_t2(sample[indices[:n1]], sample[indices[n1:]])

    else:  # Full permutation test
        combinations = itertools.combinations(indices, n1)
        dist = np.empty(int(scipy.special.comb(n, n1)))
        for i, comb in enumerate(combinations):
            sample1_i = np.asarray(comb)  # Comb is a selection of n1 indices
            sample2_i = np.setdiff1d(indices, sample1_i)  # Remaining n2 ind.
            sample1, sample2 = sample[sample1_i], sample[sample2_i]
            dist[i] = hotelling_t2(sample1, sample2)

    p_value = np.sum(dist > t2_0) / len(dist)

    if return_dist:
        return t2_0, p_value, dist

    return t2_0, p_value



"""Metrics, norms and related utilities."""

from ._angular import angular_distance
from ._fisher_rao import (
    _fisher_rao_warping_distance,
    fisher_rao_amplitude_distance,
    fisher_rao_distance,
    fisher_rao_phase_distance,
)
from ._lp_distances import (
    LpDistance,
    l1_distance,
    l2_distance,
    linf_distance,
    lp_distance,
)
from ._lp_norms import LpNorm, l1_norm, l2_norm, linf_norm, lp_norm
from ._typing import PRECOMPUTED, Metric, Norm
from ._utils import (
    NormInducedMetric,
    PairwiseMetric,
    TransformationMetric,
    pairwise_metric_optimization,
)



from ._basis import Basis
from ._bspline import BSpline
from ._coefficients_transformer import CoefficientsTransformer
from ._constant import Constant
from ._fdatabasis import FDataBasis, FDataBasisDType
from ._finite_element import FiniteElement
from ._fourier import Fourier
from ._monomial import Monomial
from ._tensor_basis import Tensor
from ._vector_basis import VectorValued



"""
Basis representation
====================

In this section, we will introduce the basis representation of
functional data. This is a very useful representation for functions that
belong (or can be reasonably projected) to the space spanned by a finite set
of basis functions.

.. Disable isort
    isort:skip_file

"""

# Author: Carlos Ramos Carreño
# License: MIT
#
# sphinx_gallery_thumbnail_number = 7

##############################################################################
# Functions and vector spaces
# ---------------------------
#
# Functions, which are the objects of study of :term:`FDA`, can be added and
# multiplied by scalars, and these operations verify the necessary properties
# to consider these functions as vectors in a vector space.
#
# The :class:`~skfda.representation.grid.FDataGrid` objects that are used to
# represent functional observations in scikit-fda also support these
# operations.

##############################################################################
# In order to show the vector operations, we create two FDatagrids with
# two functions each,
# :math:`\mathbf{X}_1 = \{x_{1i}: \mathbb{R} \to \mathbb{R}\}, i=1,2` and
# :math:`\mathbf{X}_2 = \{x_{2i}: \mathbb{R} \to \mathbb{R}\}, i=1,2`,
# and plot them.

import numpy as np
import skfda
import matplotlib.pyplot as plt

fig, axes = plt.subplots(1, 2, figsize=(8, 3))

t = np.linspace(0, 1, 100)

fd = skfda.FDataGrid(
    data_matrix=[
        np.sin(6 * t),  # First function
        2 * t,          # Second function
    ],
    grid_points=t,
)

fd.plot(axes=axes[0])
axes[0].set_title(r"$\mathbf{X}_1$")

fd2 = skfda.FDataGrid(
    data_matrix=[
        3 * t**2,              # First function
        np.log(t + 0.1),       # Second function
    ],
    grid_points=t,
)

fd2.plot(axes=axes[1])
axes[1].set_title(r"$\mathbf{X}_2$")

plt.show()

##############################################################################
# Functions can be multiplied by an scalar. This only changes the scale of
# the functions, but not their shape. Note that all the functions in the
# dataset are affected.
#
# It is also possible to add two functions together. If you do that with
# two :class:`~skfda.representation.grid.FDataGrid` objects with the same
# length, the corresponding functions will be added.

fig, axes = plt.subplots(1, 2, figsize=(8, 3))

scalar_mul = 3 * fd

scalar_mul.plot(axes=axes[0])
axes[0].set_title(r"$3 \mathbf{X}_1$")

fd_sum = fd + fd2

fd_sum.plot(axes=axes[1])
axes[1].set_title(r"$\mathbf{X}_1 + \mathbf{X}_2$")

plt.show()

##############################################################################
# Infinite (Schauder) basis
# -------------------------
#
# Some functional topological vector spaces admit a Schauder basis. This is
# a sequence of functions :math:`\Phi = \{\phi_i\}_{i=1}^{\infty}` so that
# for every function :math:`x` in the space exists a sequence of scalars
# :math:`\{a_i\}_{i=1}^{\infty}` such that
#
# .. math::
#     x(t) = \sum_{i=1}^{\infty} a_i \phi_i(t)
#
# where the convergence of this series is with respect to the vector space
# topology.
#
# If you know that your functions of interest belong to one of these vector
# spaces, it may be interesting to express your functions in a basis.
# As computers have limited memory and computation resources, it is not
# possible to obtain the infinite basis expansion. Instead, one typically
# truncates the expansion to a few basis functions, which are enough to
# approximate your observations with a certain degree of accuracy. This
# truncation also has the effect of smoothing the data, as less important
# variations, such as noise, are eliminated in the process. Moreover, as basis
# are truncated, the vector space generated by the truncated set of basis
# functions is different to the original space, and also different between
# different basis families. Thus, the choice of basis matters, even if
# originally they would have generated the same space.
#
# In scikit-fda, functions expressed as a basis expansion can be represented
# using the class :class:`~skfda.representation.basis.FDataBasis`. The main
# attributes of objects of this class are ``basis``, an object representing a
# basis family of functions, and ``coefficients``, a matrix with the scalar
# coefficients of the functions in the basis.

##############################################################################
# As an example, we can create the following function, which is expressed in
# a truncated monomial basis (and thus it is a polynomial):
#
# .. math::
#     x(t) = 3 + 2t - 4t^2 + t^3

basis = skfda.representation.basis.Monomial(
    n_basis=4,
    domain_range=(-10, 10),
)

fd_basis = skfda.FDataBasis(
    basis=basis,
    coefficients=[
        [3, 2, -4, 1],  # First (and unique) observation
    ],
)

fd_basis.plot()
plt.show()

##############################################################################
# Conversion between FDataGrid and FDataBasis
# -------------------------------------------
#
# It is possible to convert between functions in discretized form (class
# :class:`~skfda.representation.grid.FDataGrid`) and basis expansion form (
# class :class:`~skfda.representation.basis.FDataBasis`). In order to convert
# :class:`~skfda.representation.grid.FDataGrid` objects to a basis
# representation you will need to call the method ``to_basis``, passing the
# desired basis as an argument. The functions will then be projected to the
# functional basis, solving a least squares problem in order to find the
# optimal coefficients of the expansion. In order to convert a
# :class:`~skfda.representation.basis.FDataBasis` to a discretized
# representation you should call the method ``to_grid``. This method evaluates
# the functions in a grid that can be supplied as an argument in order to
# obtain the values of the discretized representation.

##############################################################################
# We now can see how the number of basis functions affect the basis expansion
# representation of a few observations taken from a real-world dataset. You
# can see that as more basis functions are used, the basis representation
# provides a better representation of the real data.

max_basis = 9

X, y = skfda.datasets.fetch_phoneme(return_X_y=True)

# Select only the first 5 samples
X = X[:5]

X.plot()

fig, axes = plt.subplots(nrows=3, ncols=3)

for n_basis in range(1, max_basis + 1):
    basis = skfda.representation.basis.Monomial(n_basis=n_basis)
    X_basis = X.to_basis(basis)

    ax = axes.ravel()[n_basis - 1]
    fig = X_basis.plot(axes=ax)
    ax.set_title(f"{n_basis} basis functions")

fig.tight_layout()
plt.show()

##############################################################################
# List of available basis functions
# ---------------------------------
#
# In this section we will provide a list of the available basis in scikit-fda.
# As explained before, the basis family is important when the basis expansion
# is truncated (which always happens in order to represent it in a computer).
# Thus, it is recommended to take a look at the available basis in order to
# pick one that provides the best representation of the original data.

##############################################################################
# First we will load a dataset to test the basis representations.

X, y = skfda.datasets.fetch_phoneme(return_X_y=True)

# Select only the first 5 samples
X = X[:5]

X.plot()
plt.show()

##############################################################################
# Monomial basis
# ^^^^^^^^^^^^^^
#
# The monomial basis (class :class:`~skfda.representation.basis.Monomial`) is
# probably one of the simpler and more well-known basis
# of functions. Often Taylor and McLaurin series are explained in the very
# first courses of Science and Engineering degrees, and students are familiar
# with polynomials since much before. Thus, the monomial basis is useful for
# teaching purposes (and that is why we have used it in the examples). It is
# also very useful for testing purposes, as it easy to manually derive the
# expected results of operations involving this basis.
#
# As a basis for functional data analysis, however, it has several issues that
# usually make preferrable to use other basis instead. First, the usual basis
# :math:`\{1, t, t^2, t^3, \ldots\}` is not orthogonal under the standard
# inner product in :math:`L^2`, that is :math:`\langle x_1, x_2 \rangle =
# \int_{\mathcal{T}} x_1(t) x_2(t) dt`. This inhibits some
# performance optimizations that are available for operations that require
# inner products. It is possible to find an orthogonal basis of polynomials,
# but it will not be as easy to understand, losing many of its advantages.
# Another problems with this basis are the necessity of a large
# number of basis functions to express local features, the bad behaviour at
# the extremes of the function and the fact that the derivatives of the basis
# expansion are not good approximations of the derivatives of the original
# data, as high order polynomials tend to have very large oscillations.

##############################################################################
# Here we show the first five elements of the monomial basis.

basis = skfda.representation.basis.Monomial(n_basis=5)
basis.plot()
plt.show()

##############################################################################
# We now show how the previous observations are represented using the first
# five elements of this basis.

X_basis = X.to_basis(basis)
X_basis.plot()
plt.show()

##############################################################################
# Fourier basis
# ^^^^^^^^^^^^^^
#
# Probably the second most well known series expansion for staticians,
# engineers, physicists and mathematicians is the Fourier series. The Fourier
# basis (class :class:`~skfda.representation.basis.Fourier`) consist on a
# constant term plus sines and cosines of varying frequency,
# all of them normalized to unit (:math:`L^2`) norm.
# This basis is a good choice for periodic functions (as a function
# expressed in this basis has the same value at the beginning and at the end
# of its domain interval if it has the same lenght as the period
# :math:`\omega`. Moreover, in this case the functions are orthonormal (that
# is why the basis used are normalized).
#
# This basis is specially indicated for functions without strong local
# features and with almost the same order of curvature everywhere, as
# otherwise the expansion require again a large number of basis to represent
# those details.

##############################################################################
# Here we show the first five elements of a Fourier basis.

basis = skfda.representation.basis.Fourier(n_basis=5)
basis.plot()
plt.show()

##############################################################################
# We now show how the previous observations are represented using the first
# five elements of this basis.

X_basis = X.to_basis(basis)
X_basis.plot()
plt.show()

##############################################################################
# B-spline basis
# ^^^^^^^^^^^^^^
#
# Splines are a family of functions that has taken importance with the advent
# of the modern computers, and nowadays are well known for a lot of engineers
# and designers. Esentially, they are piecewise polynomials that join smoothly
# at the separation points (usually called knots). Thus, both polynomials
# and piecewise linear functions are included in this family. Given a set of
# knots, a B-spline basis (class :class:`~skfda.representation.basis.BSpline`)
# of a given order can be used to express every spline of the same order that
# uses the same knots.
#
# This basis is a very powerful basis, as the knots can be adjusted to be able
# to express local features, and it is even possible to create points where
# the functions are not necessarily  smooth or continuous by placing several
# knots together. Also the elements of the basis have the compact support
# property, which allows more efficient computations. Thus, this basis is
# indicated for non-periodic functions or functions with local features or with
# different orders of curvature along their domain.

##############################################################################
# Here we show the first five elements of a B-spline basis.

basis = skfda.representation.basis.BSpline(n_basis=5)
basis.plot()
plt.show()

##############################################################################
# We now show how the previous observations are represented using the first
# five elements of this basis.

X_basis = X.to_basis(basis)
X_basis.plot()
plt.show()

##############################################################################
# Constant basis
# ^^^^^^^^^^^^^^
#
# Sometimes it is useful to consider the basis whose only function is the
# constant one. In particular, using this basis we can view scalar values
# as functional observations, which can be used to combine multivariate
# and functional data in the same model.

##############################################################################
# Tensor product basis
# ^^^^^^^^^^^^^^^^^^^^
#
# The previously explained bases are useful for data that comes in the form
# of curves, that is, functions :math:`\{f_i: \mathbb{R} \to
# \mathbb{R}\}_{i=1}^N`. However, scikit-fda allows also the representation
# of surfaces or functions in higher dimensions. In this case it is even more
# useful to be able to represent them using basis expansions, as the number
# of parameters in the discretized representation grows as the product of the
# grid points in each dimension of the domain.
#
# The tensor product basis (class :class:`~skfda.representation.basis.Tensor`)
# allows the construction of basis for these higher dimensional functions as
# tensor products of :math:`\mathbb{R} \to \mathbb{R}` basis.

##############################################################################
# As an example, we can import the digits datasets of scikit-learn, which are
# surfaces, and convert it to a basis expansion. Note that we use different
# basis for the different continuous parameters of the function in order to
# show how it works, although it probably makes no sense in this particular
# example.

from sklearn.datasets import load_digits

X, y = load_digits(return_X_y=True)
X = X.reshape(-1, 8, 8)

fd = skfda.FDataGrid(X)

basis = skfda.representation.basis.Tensor([
    skfda.representation.basis.Fourier(  # X axis
        n_basis=5,
        domain_range=fd.domain_range[0],
    ),
    skfda.representation.basis.BSpline(  # Y axis
        n_basis=6,
        domain_range=fd.domain_range[1],
    ),
])

fd_basis = fd.to_basis(basis)

# We only plot the first function
fd_basis[0].plot()
plt.show()

##############################################################################
# Finite element basis
# ^^^^^^^^^^^^^^^^^^^^
#
# A finite element basis (class
# :class:`~skfda.representation.basis.FiniteElement`) is a basis used in the
# finite element method (FEM). In order to instantiate a basis, it is
# necessary to pass a set of vertices and a set of simplices, or cells, that
# join them, conforming a grid. The basis elements are then functions that
# are one at exactly one of these vertices and zero in the rest of them.
#
# The advantage of this basis for higher dimensional functions is that one can
# have more control of the basis, placing more vertices in regions with
# interesting behaviour, such as local features and less elsewhere.

##############################################################################
# Here we show an example where the digits dataset of scikit-learn is
# expressed in the finite element basis. First we create the vertices and
# simplices that we will use and we plot them.

vertices = np.array([
    (0, 0),
    (0, 1),
    (1, 0),
    (1, 1),
    (0.25, 0.5),
    (0.5, 0.25),
    (0.5, 0.75),
    (0.75, 0.5),
    (0.5, 0.5),
])

cells = np.array([
    (0, 1, 4),
    (0, 2, 5),
    (1, 3, 6),
    (2, 3, 7),
    (0, 4, 5),
    (1, 4, 6),
    (2, 5, 7),
    (3, 6, 7),
    (4, 5, 8),
    (4, 6, 8),
    (5, 7, 8),
    (6, 7, 8),
])

plt.triplot(vertices[:, 0], vertices[:, 1], cells)
plt.show()

##############################################################################
# We now represent the digits dataset in this basis.

basis = skfda.representation.basis.FiniteElement(
    vertices=vertices,
    cells=cells,
)

fd_basis = fd.to_basis(basis)

# We only plot the first function
fd_basis[0].plot()
plt.show()

##############################################################################
# Vector-valued basis
# ^^^^^^^^^^^^^^^^^^^
#
# With the aforementioned bases, one could express
# :math:`\mathbb{R}^p \to \mathbb{R}` functions. In order to express vector
# valued functions as a basis expansion, one just need to express each
# coordinate function as a basis expansion and multiply it by the
# corresponding unitary vector in the coordinate direction, adding finally all
# of them together.
#
# The vector-valued basis (:class:`~skfda.representation.basis.VectorValued`)
# allows the representation of vector-valued functions doing just that.

##############################################################################
# As an example, consider the Canadian Weather dataset, including both
# temperature and precipitation data as coordinate functions, and plotted
# below.

X, y = skfda.datasets.fetch_weather(return_X_y=True)

X.plot()
plt.show()

##############################################################################
# We will express this dataset as a basis expansion. Temperatures
# are now expressed in a Fourier basis, while we express precipitations as
# B-splines.

basis = skfda.representation.basis.VectorValued([
    skfda.representation.basis.Fourier(  # First coordinate function
        n_basis=5,
        domain_range=X.domain_range,
    ),
    skfda.representation.basis.BSpline(  # Second coordinate function
        n_basis=10,
        domain_range=X.domain_range,
    ),
])

X_basis = X.to_basis(basis)
X_basis.plot()
plt.show()



from ._fisher_rao import _fisher_rao_warping_mean, fisher_rao_karcher_mean
from ._stats import (
    cov,
    depth_based_median,
    geometric_median,
    gmean,
    mean,
    modified_epigraph_index,
    trim_mean,
    var,
)



from . import multivariate
from ._depth import (IntegratedDepth,
                     ModifiedBandDepth,
                     BandDepth)
from .multivariate import Depth, Outlyingness, OutlyingnessBasedDepth
